"""Unit tests for Label class."""

from pytest_cases.case_parametrizer_new import parametrize_with_cases

from uppaalpy.classes.class_tests.test_label_cases import (
    CaseLabelFromElement,
    CaseLabelInit,
    CaseLabelToElement,
    CaseUpdateLabelInit,
    CaseUpdateResets,
)
from uppaalpy.classes.simplethings import Label, UpdateLabel


class TestLabel:
    @parametrize_with_cases("kind, val, pos", cases=CaseLabelInit)
    def test_init(self, kind, val, pos):
        Label(kind, val, pos)

    @parametrize_with_cases("element", cases=CaseLabelFromElement)
    def test_from_element(self, element):
        Label.from_element(element)

    @parametrize_with_cases("element", cases=CaseLabelToElement)
    def test_to_element(self, element):
        e = Label.from_element(element)
        element2 = e.to_element()
        assert element.tag == element2.tag
        assert element.text == element2.text
        assert element.get("x") == element2.get("x")
        assert element.get("y") == element2.get("y")


class TestUpdateLabel:
    @parametrize_with_cases("kind, val, pos, ctx, updates", cases=CaseUpdateLabelInit)
    def test_init(self, kind, val, pos, ctx, updates):
        l = UpdateLabel(kind, val, pos, ctx, updates)

        assert l.value == val
        assert l.kind == kind
        assert l.pos == pos
        if updates is not None:
            assert l.updates == updates

    @parametrize_with_cases("label, res", cases=CaseUpdateResets)
    def test_resets(self, label, res):
        assert label.get_resets() == res



"""Test path realizability."""

import pytest
from uppaalpy.path_tests.helpers import *


@pytest.mark.parametrize("fp", reachable_nta())
def test_reachable_path(fp):
    """Test realizable paths."""
    nta = read_file(fp)
    path = get_path_from_query_comment(nta)
    assert p.path_realizable(path, True)[0] == True


@pytest.mark.parametrize("fp", not_reachable_nta())
def test_not_reachable_path(fp):
    """Test not realizable paths."""
    nta = read_file(fp)
    path = get_path_from_query_comment(nta)
    assert p.path_realizable(path)[0] == False


@pytest.mark.parametrize("fp", reachable_nta())
def test_reachable_path_with_ge_zero_clocks(fp):
    """Test path realizability for all segments of the path."""
    nta = read_file(fp)
    path = get_path_from_query_comment(nta)
    # path: l - t - l - t - l... for l:location and t:transition
    for i in range(0, len(path), 2):
        for j in range(0, i, 2):
            assert (
                p.path_realizable_with_initial_valuation(
                    path[j:i+1], True, icv_constants={}
                )[0]
                == True
            )



"""Unit tests for expressions."""
from pytest_cases.case_parametrizer_new import parametrize_with_cases

from uppaalpy.classes.class_tests.test_expr_cases import (
    CaseClockConstraintClock,
    CaseClockResetInit,
    CaseConstraintHandle,
    CaseConstraintParse,
    CaseExprTokenize,
    CaseUpdateExprJoin,
    CaseUpdateExprSplit,
    CaseUpdateHandle,
    CaseUpdateInit,
)
from uppaalpy.classes.expr import (
    ClockConstraintExpression,
    ClockResetExpression,
    ConstraintExpression,
    Expression,
    UpdateExpression,
)


class TestExpr:
    @parametrize_with_cases("string, res", cases=CaseExprTokenize)
    def test_tokenize(self, string, res):
        r = Expression.tokenize(string)
        assert len(r) == 3
        if res is not None:
            for result, answer in zip(r, res):
                assert result == answer


class TestUpdateExpr:
    @parametrize_with_cases("string, ctx", cases=CaseUpdateInit)
    def test_init(self, string, ctx):
        UpdateExpression(string, ctx)

    @parametrize_with_cases("string, ctx", cases=CaseUpdateInit)
    def test_to_str(self, string, ctx):
        r = UpdateExpression(string, ctx)
        assert string == r.to_string()

    @parametrize_with_cases("string, res", cases=CaseUpdateExprSplit)
    def test_split(self, string, res):
        r = UpdateExpression.split_into_simple(string)
        for i, subexprstr in enumerate(r):
            assert "," not in subexprstr
            if res is not None:
                assert subexprstr == res[i]

    @parametrize_with_cases("strings, res", cases=CaseUpdateExprJoin)
    def test_join_str(self, strings, res):
        r = UpdateExpression.join_strings(strings)
        assert r == res

    @parametrize_with_cases("expr, ctx, res, diff", cases=CaseUpdateHandle)
    def test_handle_update(self, expr, ctx, res, diff):
        old = ctx.get_val(expr.lhs)
        expr.handle_update(ctx)
        new = ctx.get_val(expr.lhs)

        if res is None:
            assert new - old == diff

        else:
            assert new == res


class TestClockReset:
    @parametrize_with_cases("string, ctx", cases=CaseClockResetInit)
    def test_init(self, string, ctx):
        e = ClockResetExpression(string, ctx)
        assert e.clock == e.lhs
        assert ctx.is_clock(e.clock)


class TestConstraintExpr:
    @parametrize_with_cases(
        "string, ctx, is_clock_constraint", cases=CaseConstraintParse
    )
    def test_parse(self, string, ctx, is_clock_constraint):
        expr = ConstraintExpression.parse_expr(string, ctx)
        assert isinstance(expr, ClockConstraintExpression) == is_clock_constraint

    @parametrize_with_cases("string, ctx, res", cases=CaseConstraintHandle)
    def test_handle(self, string, ctx, res):
        expr = ConstraintExpression(string, ctx)
        assert expr.handle_constraint(ctx) == res


class TestClockConstraint:
    @parametrize_with_cases(
        "string, ctx, res_clock, res_thres", cases=CaseClockConstraintClock
    )
    def test_clock_and_thres(self, string, ctx, res_clock, res_thres):
        expr = ClockConstraintExpression(string, ctx)
        assert expr.clocks == res_clock
        assert expr.threshold == res_thres



from uppaalpy.classes.context import Context


class DataContext:
    @classmethod
    def ctx(cls):
        return Context(
            {"c1", "c2", "c"},
            {"x": -10, "y": 0, "constant3": 3},
            {"i": -1, "j": 4, "k": 10},
        )



from uppaalpy.classes.class_tests.test_context_cases import DataContext
from uppaalpy.classes.constraint_patcher import ConstraintInsert, ConstraintRemove, ConstraintUpdate
from uppaalpy.classes.expr import ClockConstraintExpression
from uppaalpy.classes.simplethings import ConstraintLabel


class CaseCRInit:
    class C(DataContext):
        pass

    def case_01(self):
        return ClockConstraintExpression("c > 15", self.C.ctx()), False

    def case_02(self):
        return ClockConstraintExpression("c > 15", self.C.ctx()), True

    def case_03(self):
        return ClockConstraintExpression("c > 15", self.C.ctx()), None


class CaseRFind:
    class C(DataContext):
        pass

    def case_01(self):
        return ClockConstraintExpression("c == 15", self.C.ctx()), ["c == 15"], 0

    def case_02(self):
        return (
            ClockConstraintExpression("c == 15", self.C.ctx()),
            ["c == 15", "c1 > 13"],
            0,
        )

    def case_03(self):
        return (
            ClockConstraintExpression("c == 15", self.C.ctx()),
            ["c1 > 13", "c == 15"],
            1,
        )

    def case_04(self):
        return (
            ClockConstraintExpression("c == 15", self.C.ctx()),
            ["c1 > 13", "c2 == 15"],
            -1,
        )

    def case_05(self):
        return ClockConstraintExpression("c == 15", self.C.ctx()), [], -1


class CaseRPatch:
    class C(DataContext):
        pass

    def case_01(self):
        lines = [
            '		<location id="id0" x="0" y="0">\n',
            '			<label kind="invariant" x="18" y="-34">c &lt; 5 &amp;&amp; i &lt; 5</label>\n',
            "		</location>\n",
        ]
        res = [
            '		<location id="id0" x="0" y="0">\n',
            '			<label kind="invariant" x="18" y="-34">i &lt; 5</label>\n',
            "		</location>\n",
        ]
        expr = ClockConstraintExpression("c < 5", self.C.ctx())
        cr = ConstraintRemove(expr, False)

        return cr, lines, 1, res

    def case_02(self):
        lines = [
            '		<location id="id0" x="0" y="0">\n',
            '			<label kind="invariant" x="18" y="-34">c &lt; 5</label>\n',
            "		</location>\n",
        ]
        res = [
            '		<location id="id0" x="0" y="0">\n',
            "		</location>\n",
        ]
        expr = ClockConstraintExpression("c < 5", self.C.ctx())
        cr = ConstraintRemove(expr, True)

        return cr, lines, 1, res

    def case_03(self):
        lines = [
            '		<location id="id0" x="0" y="0">\n',
            '			<label kind="name" x="18" y="-34">location0</label>\n',
            '			<label kind="invariant" x="18" y="-34">c &lt; 5</label>\n',
            "		</location>\n",
        ]

        res = [
            '		<location id="id0" x="0" y="0">\n',
            '			<label kind="name" x="18" y="-34">location0</label>\n',
            "		</location>\n",
        ]

        expr = ClockConstraintExpression("c < 5", self.C.ctx())
        cr = ConstraintRemove(expr, True)

        return cr, lines, 2, res

    def case_04(self):
        lines = [
            '		<location id="id0" x="0" y="0">\n',
            '			<label kind="name" x="18" y="-34">location0</label>\n',
            '			<label kind="invariant" x="18" y="-34">c &lt; 5</label>\n',
            '			<label kind="exponentialrate" x="18" y="-34">foo</label>\n',
            "		</location>\n",
        ]
        res = [
            '		<location id="id0" x="0" y="0">\n',
            '			<label kind="name" x="18" y="-34">location0</label>\n',
            '			<label kind="exponentialrate" x="18" y="-34">foo</label>\n',
            "		</location>\n",
        ]

        expr = ClockConstraintExpression("c < 5", self.C.ctx())
        cr = ConstraintRemove(expr, True)

        return cr, lines, 2, res

    def case_05(self):
        lines = [
            "		<transition>\n",
            '			<source ref="id0"/>\n',
            '			<source ref="id1"/>\n',
            '			<label kind="guard" x="18" y="-34">5 == c</label>\n',
            "		</transition>\n",
        ]
        res = [
            "		<transition>\n",
            '			<source ref="id0"/>\n',
            '			<source ref="id1"/>\n',
            "		</transition>\n",
        ]
        expr = ClockConstraintExpression("5 == c", self.C.ctx())
        cr = ConstraintRemove(expr, True)

        return cr, lines, 3, res


class CaseIinit:
    class C(DataContext):
        pass

    def case_01(self):
        return ClockConstraintExpression("c > 15", self.C.ctx()), None


class CaseIPatch:
    class C(DataContext):
        pass

    def case_01(self):
        lines = [
            '		<location id="id0" x="0" y="0">\n',
            "		</location>\n",
        ]

        lines_expected = [
            '		<location id="id0" x="0" y="0">\n',
            '			<label kind="invariant" x="18" y="-34">c1 - c &lt;= 5</label>\n',
            "		</location>\n",
        ]

        expr = ClockConstraintExpression("c1 - c <= 5", self.C.ctx())
        invariant = ConstraintLabel("invariant", "", (18, -34), self.C.ctx(), [expr])
        ci = ConstraintInsert(expr, invariant)

        return ci, lines, lines_expected, 0, 0


    def case_02(self):
        lines = [
            '		<location id="id0" x="0" y="0">\n',
            '			<label kind="invariant" x="18" y="-34">x - y &lt;= 5</label>\n',
            "		</location>\n",
        ]

        lines_expected = [
            '		<location id="id0" x="0" y="0">\n',
            '			<label kind="invariant" x="18" y="-34">x - y &lt;= 5 &amp;&amp; c == 3</label>\n',
            "		</location>\n",
        ]

        expr = ClockConstraintExpression("c == 3", self.C.ctx())
        ci = ConstraintInsert(expr)

        return ci, lines, lines_expected, 1, 0

    def case_03(self):
        lines = [
            "		<transition>\n",
            '			<source ref="id0"/>\n',
            '			<source ref="id1"/>\n',
            "		</transition>\n",
        ]

        lines_expected = [
            "		<transition>\n",
            '			<source ref="id0"/>\n',
            '			<source ref="id1"/>\n',
            '			<label kind="guard" x="18" y="-34">c == 5</label>\n',
            "		</transition>\n",
        ]

        expr = ClockConstraintExpression("c == 5", self.C.ctx())
        guard = ConstraintLabel("guard", "", (18, -34), self.C.ctx(), [expr])
        ci = ConstraintInsert(expr, guard)

        return ci, lines, lines_expected, 2, 0

    def case_04(self):
        lines = [
            "		<transition>\n",
            '			<source ref="id0"/>\n',
            '			<source ref="id1"/>\n',
            '			<label kind="guard" x="18" y="-34">clock2 == 5</label>\n',
            '			<label kind="synchronisation" x="18" y="-34">foo</label>\n',
            "		</transition>\n",
        ]

        lines_expected = [
            "		<transition>\n",
            '			<source ref="id0"/>\n',
            '			<source ref="id1"/>\n',
            '			<label kind="guard" x="18" y="-34">clock2 == 5 &amp;&amp; c == 5</label>\n',
            '			<label kind="synchronisation" x="18" y="-34">foo</label>\n',
            "		</transition>\n",
        ]

        expr = ClockConstraintExpression("c == 5", self.C.ctx())
        ci = ConstraintInsert(expr)

        return ci, lines, lines_expected, 3, 0

class CaseUInit:
    class C(DataContext):
        pass

    def case_01(self):
        c = ClockConstraintExpression("c == 3", self.C.ctx())
        cu = ConstraintUpdate(c, "11")

        return c, cu, "3", "11"

class CaseUFind:
    class C(DataContext):
        pass
    c = ClockConstraintExpression("c > 10", C.ctx())
    cu = ConstraintUpdate(c, "13")

    def case_01(self):
        return self.cu, ["c &gt; 10"], 0

    def case_02(self):
        return self.cu, ["c1 == 10", "c &gt; 10"], 1

    def case_03(self):
        return self.cu, ["c1 &gt; 13"], None

class CaseUPatch:
    class C(DataContext):
        pass

    def case_01(self):
        lines = [
            '		<location id="id0" x="0" y="0">\n',
            '			<label kind="invariant" x="18" y="-34">c &lt; 5 &amp;&amp; y &lt; 5</label>\n',
            "		</location>\n",
        ]

        lines_expected = [
            '		<location id="id0" x="0" y="0">\n',
            '			<label kind="invariant" x="18" y="-34">c &lt; 10 &amp;&amp; y &lt; 5</label>\n',
            "		</location>\n",
        ]

        c = ClockConstraintExpression("c < 5", self.C.ctx())
        cu = ConstraintUpdate(c, "10")

        return cu, lines, 1, lines_expected




"""The definition of Transition class resides here."""

from typing import List, Optional, Type

import lxml.etree as ET

from uppaalpy.classes import templates as t
from uppaalpy.classes.context import Context
from uppaalpy.classes.expr import ConstraintExpression
from uppaalpy.classes.simplethings import ConstraintLabel, Label, UpdateLabel


class Transition:
    """Class for edges of the TA.

    Many of the attributes are optional. Non existent ones are simply stored as
    None.

    Attributes:
        source: String of the form "idX". References locations or branchpoints.
        target: String of the form "idX". References locations or branchpoints.
        select: Label object with kind 'select'. See UPPAAL documentation.
        guard: ConstraintLabel object with kind 'guard'. See UPPAAL
            documentation.
        synchronisation: Label object with kind 'synchronisation'.
            See UPPAAL documentation.
        assignment: Update Label object with kind 'assignment'. See UPPAAL
            documentation.
        testcode: Label object with kind 'testcode'. See UPPAAL documentation.
        probability: Label object with kind 'probability'. See UPPAAL
            documentation.
        comments: Label object with kind 'comments'. See UPPAAL
            documentation.
        nails: List of Nail objectsAny, .
    """

    def __init__(self, **kwargs) -> None:
        """Construct a Transition object from keyword args.

        All arguments except source and target are optional.

        Args:
            source: String, id of the source location.
            target: String, id of the target location.
            select: Label with kind "select". See UPPAAL documentation.
            guard: ConstraintLabel object.
            synchronisation: Label with kind "synchronisation".
            assignment: UpdateLabel object.
            testcode: Label with kind "testcode".
            probability: Label with kind "probability".
            comments: Label...
            nails: List of Nail objects.
            template: The parent template. Set by TAGraph.
        """
        self.source = kwargs["source"]  # type: str
        self.target = kwargs["target"]  # type: str
        self.select = kwargs.get("select")  # type: Optional[Label]
        self.guard = kwargs.get("guard")  # type: Optional[ConstraintLabel]
        self.synchronisation = kwargs.get("synchronisation")  # type: Optional[Label]
        self.assignment = kwargs.get("assignment")  # type: Optional[UpdateLabel]
        self.testcode = kwargs.get("testcode")  # type: Optional[Label]
        self.probability = kwargs.get("probability")  # type: Optional[Label]
        self.comments = kwargs.get("comments")  # type: Optional[Label]
        self.nails = (
            kwargs.get("nails") if kwargs.get("nails") is not None else []
        )  # type: Optional[List[Nail]]
        self.template: Optional[t.Template] = None  # Optional[Template]

    @classmethod
    def from_element(cls: Type["Transition"], et, ctx: Context) -> "Transition":
        """Construct a Transition from an Element object, and return it."""
        kw = {}
        kw["source"] = et.find("source").get("ref")
        kw["target"] = et.find("target").get("ref")

        for label in et.iter("label"):
            l_kind = label.get("kind")
            label_obj = Label.from_element(label)
            if l_kind == "guard":
                label_obj = ConstraintLabel.from_label(label_obj, ctx)
            elif l_kind == "assignment":
                label_obj = UpdateLabel.from_label(label_obj, ctx)
            kw[l_kind] = label_obj

        kw["nails"] = [
            Nail(int(nail.get("x")), int(nail.get("y"))) for nail in et.iter("nail")
        ]

        return cls(**kw)

    def to_element(self):
        """Convert this object to an Element."""
        element = ET.Element("transition")
        element.append(ET.Element("source", attrib={"ref": self.source}))
        element.append(ET.Element("target", attrib={"ref": self.target}))
        if self.select is not None:
            element.append(self.select.to_element())
        if self.guard is not None:
            element.append(self.guard.to_element())
        if self.synchronisation is not None:
            element.append(self.synchronisation.to_element())
        if self.assignment is not None:
            element.append(self.assignment.to_element())
        if self.testcode is not None:
            element.append(self.testcode.to_element())
        if self.probability is not None:
            element.append(self.probability.to_element())
        if self.comments is not None:
            element.append(self.comments.to_element())
        for nail in self.nails:
            element.append(nail.to_element())
        return element

    def get_constraints(self) -> List[ConstraintExpression]:
        """Return a list of simple constraints on this transition."""
        if self.guard is not None:
            return self.guard.constraints
        else:
            return []

    def get_constraint_label(self) -> Optional[ConstraintLabel]:
        """Return the guard label."""
        return self.guard

    def create_constraint_label(self, exp: ConstraintExpression, ctx: Context) -> None:
        """Create a guard label with the given expression."""
        t = self.template
        t_name = t.name.name
        src, dst = (t_name, self.source), (t_name, self.target)
        slx, sly = t.graph.nodes[src]["obj"].pos
        dlx, dly = t.graph.nodes[dst]["obj"].pos
        guard_pos = (slx + dlx, sly + dly)

        self.guard = ConstraintLabel("guard", "", guard_pos, ctx, [exp])

    def remove_constraint_label(self) -> None:
        """Remove guard."""
        self.guard = None


class Nail:
    """Class for storing 'nails' on the edges of the TA.

    Attributes:
        pos: Pair of ints.
    """

    def __init__(self, x: int, y: int) -> None:
        """Construct Nail from an int pair.

        Args:
            x: Int, the x comp. of the position of the Nail.
            y: Int, the y comp. of the position of the Nail.
        """
        self.pos = x, y

    @classmethod
    def from_element(cls: Type["Nail"], et) -> "Nail":
        """Construct Nail from an Element."""
        return cls(int(et.get("x")), int(et.get("y")))

    def to_element(self):
        """Construct an element from Nail object."""
        return ET.Element("nail", attrib={"x": str(self.pos[0]), "y": str(self.pos[1])})



"""Class definitions for fast constraint changes."""

from abc import ABCMeta, abstractmethod
from copy import copy
from typing import List, Optional, Union, cast

from uppaalpy.classes import nodes as n
from uppaalpy.classes import nta
from uppaalpy.classes import templates as te
from uppaalpy.classes import transitions as tr
from uppaalpy.classes.expr import ClockConstraintExpression
from uppaalpy.classes.simplethings import ConstraintLabel


class ConstraintCache:
    """Class for line based constraint changes.

    (De)serializing with lxml and traversing/creating an Element Tree for
    small updates on constraints is wasteful. Caching updates on constraints
    by creating patches and applying these patches by linewise editing might
    be a faster alternative. However, since the xml file is not actually parsed,
    the implementation requires the file to be formatted properly. Precisely,
    each "leaf" item must be on a separate single line, and each "inner" item
    must have their "start" and "end" on a separate single line:

    example:
    <transition>
        <source ref="id1"/>
        <target ref="id0"/>
        <label kind="guard" x="289" y="-25">x &gt;= 100</label>
    </transition>
    """

    def __init__(self, nta: nta.NTA) -> None:
        """Initialize ConstraintCache.

        Attributes:
            patches: List of ConstraintPatch objects. Those patches are
                "flushed" to an output file when the user choses to do so.
            nta: The parent nta.
        """
        self.patches: List[ConstraintPatch] = []
        self.nta = nta

    def cache(self, patch: "ConstraintPatch") -> None:
        """Store a patch."""
        self.patches.append(patch)

    def _apply_single_patch(self, lines: List[str], patch: "ConstraintPatch") -> None:
        """Apply a single patch."""

        def handle_loc(i: int, loc: n.Location) -> None:
            # Find the line with the relevant location.
            loc_string = '<location id="%s"' % loc.id
            while loc_string not in lines[i]:
                i += 1
            location_line_index = i
            # If no invariant exists for this location in the file,
            # create a new line for the new invariant label. It should be
            # inserted just after the Name label, if it exists, and before
            # all the other labels.
            target_index = location_line_index

            # Find the invariant line, if it exists.
            while True:
                line = lines[i].strip()
                if line.startswith('<name x="'):
                    # Invariant comes after name.
                    target_index = i
                if line.startswith('<label kind="invariant"'):  # Invariant found.
                    target_index = i
                    break
                if line.startswith("</location>"):  # No invariant
                    break
                i += 1

            patch.change.patch_line(lines, target_index, location_line_index)

        def handle_trans(i: int, trans: tr.Transition) -> None:
            # Find the line with the relevant transition.
            trans_index = trans.template.graph._transitions.index(trans)
            curr_trans = -1
            while curr_trans < trans_index:
                if lines[i].strip().startswith("<transition>"):
                    curr_trans += 1
                i += 1
            transition_line_index = i - 1
            # If no guard exists for this tranisiton in the file,
            # create a new line for the new guard label. It should be
            # inserted just after the Name label, if one exists, and before
            # all the other labels.
            target_index = transition_line_index + 2  # skip source and target lines

            # Find the guard line, if it exists.
            while True:
                line = lines[i].strip()
                if line.startswith('<label kind="select"'):
                    # Guard comes after select.
                    target_index = i
                if line.startswith('<label kind="guard"'):  # Guard found.
                    target_index = i
                    break
                if line.startswith("</transition>"):  # No invariant
                    break
                i += 1

            patch.change.patch_line(lines, target_index, transition_line_index)

        template_index = self.nta.templates.index(patch.template_ref)
        i = 0
        curr_template_i = -1

        # Find the line the template starts.
        while curr_template_i < template_index:
            if lines[i].strip().startswith("<template>"):
                curr_template_i += 1
            i += 1

        # Check whether the change is on a location or a transition.
        if type(patch.obj_ref) == n.Location:
            handle_loc(i, cast(n.Location, patch.obj_ref))

        else:
            handle_trans(i, cast(tr.Transition, patch.obj_ref))

    def apply_patches(self, lines: List[str]):
        """Given a list of lines, apply changes the list."""
        for patch in self.patches:
            self._apply_single_patch(lines, patch)


class ConstraintPatch:
    """Class for capturing a change on guards and invariants.

    Currently insertion, removal, and update operations on clock constraint
    expressions are supported.
    """

    def __init__(
        self,
        template_ref: te.Template,
        change: "ConstraintChange",
        obj_ref: Union[n.Location, tr.Transition],
    ) -> None:
        """Initialize ConstraintPatch.

        obj_ref argument can be used for initializing the patch

        Args:
            template_ref: The parent template.
            change: A ConstraintChange object.
            location_ref: The parent location.
            transition_ref: The parent transition.
        """
        self.template_ref = template_ref
        self.change = change
        self.obj_ref = obj_ref


class ConstraintChange(metaclass=ABCMeta):
    """Base class for the three operations on constraint changes.

    Attributes:
        constraint: A ClockConstraintExpression object.
    """

    def __init__(self, constraint: ClockConstraintExpression) -> None:
        """Initialize class with the clock constraint expr to be changed."""
        self.constraint = constraint

    @abstractmethod
    def patch_line(self, lines: List[str], index: int, parent_index: int = -1) -> None:
        """Patch a list of lines."""
        pass


class ConstraintRemove(ConstraintChange):
    """Class for keeping track of a constraint removal."""

    def __init__(
        self, constraint: ClockConstraintExpression, remove_label: bool = False
    ) -> None:
        """Create ConstraintRemove given a simple constraint to remove."""
        super().__init__(constraint)
        self.remove_label = remove_label

    def patch_line(self, lines: List[str], index: int, parent_index: int = -1) -> None:
        """Remove a constraint by editing or deleting a line.

        Args:
            lines: List of strings for each line.
            index: Integer index of the current line. If self.remove_constraint
                is False, current line is edited. Otherwise, the line is
                deleted.
            parent_index: Not used.
        """
        parent_index  # for ignoring unused hint.
        if self.remove_label:
            lines.pop(index)

        else:
            # Edit the current line.
            constraint_line = lines[index]
            start = constraint_line.index(">") + 1  # '>' in ...y="..">
            end = constraint_line.index("<", start)  # '<' in </label>
            constraints = constraint_line[start:end].split(" &amp;&amp; ")
            constraints.pop(self._find_matching_constraint(constraints))
            lines[index] = (
                constraint_line[:start]
                + " &amp;&amp; ".join(constraints)
                + constraint_line[end:]
            )

    def _find_matching_constraint(self, constraints: List[str]) -> int:
        """Find the index of the constraint to be deleted.

        Each string is compared with the constraint to be removed.
        """
        comparison_string = self.constraint.to_string(escape=True).replace(" ", "")

        for i, c in enumerate(constraints):
            if c.replace(" ", "") == comparison_string:
                return i

        raise Exception(
            "{comp} does not match with any of the {lst}".format(
                comp=comparison_string, lst=constraints
            )
        )


class ConstraintInsert(ConstraintChange):
    """Class for keeping track of a constraint insertion."""

    def __init__(
        self,
        constraint: ClockConstraintExpression,
        newly_created: Optional[ConstraintLabel] = None,
    ) -> None:
        """Create ConstraintInsert given a simple constraint to insert.

        If a new guard/invariant is created, self.newly_created is set to the
        newly created object. Otherwise, it is None.
        """
        super().__init__(constraint)
        self.newly_created = newly_created

    def patch_line(self, lines: List[str], index: int, parent_index: int) -> None:
        """Insert a constraint by editing or inserting a line.

        Args:
            lines: List of strings for each line.
            index: Integer index of the current line. If self.created_new is
                not none, a new line after the current line is inserted for the
                new invariant/guard label. Otherwise, the new constraint is inserted
                to the current line.
            parent_index: Integer index of the parent transition/location.
                Used for indentation while inserting a new line.
        """
        if self.newly_created is not None:
            # Insert new line after the current line.
            tabs = lines[parent_index].index("<") + 1
            string = (
                tabs * "\t"
                + '<label kind="{kind}" x="{x}" y="{y}">{text}</label>\n'.format(
                    kind=self.newly_created.kind,
                    x=str(self.newly_created.pos[0]),
                    y=str(self.newly_created.pos[1]),
                    text=self.constraint.to_string(escape=True),
                )
            )
            lines.insert(index + 1, string)

        else:
            # Edit the current line.
            constraint_line = lines[index]
            start = constraint_line.index(">") + 1  # '>' in ...y="..">
            insertion_point = constraint_line.index("<", start)  # '<' in </label>
            edited_line = "{prev} &amp;&amp; {text}{rest}".format(
                prev=constraint_line[:insertion_point],
                text=self.constraint.to_string(escape=True),
                rest=constraint_line[insertion_point:],
            )
            lines[index] = edited_line


class ConstraintUpdate(ConstraintChange):
    """Class for keeping track of a constraint update."""

    def __init__(
        self, constraint: ClockConstraintExpression, new_threshold: str
    ) -> None:
        """Initialize class with the new and the old thresholds."""
        super().__init__(constraint)
        self.old = constraint.threshold
        self.new = new_threshold

    def patch_line(self, lines: List[str], index: int, parent_index: int = -1) -> None:
        """Update a constraint by editing a line.

        Args:
            lines: List of strings for each line.
            index: Integer index of the current line.
            parent_index: Not used.
        """
        parent_index
        constraint_line = lines[index]
        start = constraint_line.index(">") + 1  # '>' in ...y="..">
        end = constraint_line.index("<", start)  # '<' in </label>
        constraints = constraint_line[start:end].split(" &amp;&amp; ")
        update_index = self._find_matching_constraint(constraints)
        constraints[update_index] = constraints[update_index].replace(
            self.old, self.new
        )
        lines[index] = (
            constraint_line[:start]
            + " &amp;&amp; ".join(constraints)
            + constraint_line[end:]
        )

    def _find_matching_constraint(self, constraints: List[str]) -> int:
        """Find the index of the constraint to be updated.

        Each string is compared with the constraint to be updated.
        """
        # Get the old comparison string.
        comparison_string = (
            self.constraint.to_string(escape=True)
            .replace(" ", "")
            .replace(self.constraint.threshold, self.old)
        )

        for i, c in enumerate(constraints):
            if c.replace(" ", "") == comparison_string:
                return i

        raise Exception(
            "{comp} does not match with any of the {lst}".format(
                comp=comparison_string, lst=constraints
            )
        )

    def generate_new_constraint(self) -> ClockConstraintExpression:
        """Create a copy ClockConstraintExpression with the updated threshold."""
        res = copy(self.constraint)
        res.threshold = self.new
        return res



"""Tests for path existance check logic."""
import pytest
from uppaalpy.path_tests.helpers import *

@pytest.mark.parametrize("fp", path_exists())
def test_path_exists(fp):
    nta = read_file(fp)
    path = get_path_from_query_comment(nta)
    assert p.path_exists(path)

@pytest.mark.parametrize("fp", path_not_exists())
def test_path_not_exists(fp):
    nta = read_file(fp)
    path = get_path_from_query_comment(nta)
    assert not p.path_exists(path)



"""Check the behavior of the path analysis on open/closed NTA."""
import pytest
from uppaalpy.path_tests.helpers import *

@pytest.mark.parametrize("fp", epsilon_tests())
def test_epsilon(fp):
    nta = read_file(fp)
    path = get_path_from_query_comment(nta)
    assert p.path_realizable(path, add_epsilon=False)[0] \
            and not p.path_realizable(path, add_epsilon=True)[0]



"""Template class definition.

Each template in an NTA represents a Distinct TA "recipe".
"""
from typing import Type

import lxml.etree as ET

from uppaalpy.classes import nodes as n
from uppaalpy.classes import tagraph as g
from uppaalpy.classes import transitions as tr
from uppaalpy.classes.context import Context
from uppaalpy.classes.simplethings import Declaration, Name, Parameter


class Template:
    """Template for extended timed automaton.

    Attributes:
        name: Name object for storing name of the TA.
        parameter: Parameter object with template parameters.
        declaration: Declaration object with local declarations.
        context: Context object containing declarations and the initial
            valuations of variables.
        graph: TAGraph object with location and branchpoints
            as nodes and transitions as edges. Superclass NetworkX MultiDigraphs
            require nodes and edges to be hashable. For this reason, actual Node
            and Transition objects are 'attached' to the graph nodes and edges
            as attributes.
    """

    def __init__(self, ctx: Context) -> None:
        """Construct a Template.

        Attributes are set by the from_element method.
        """
        self.name = Name("", (0, 0))
        self.parameter = Parameter("")
        self.declaration = Declaration("")
        self.context = ctx
        self.graph = g.TAGraph(self)

    @classmethod
    def from_element(cls: Type["Template"], et, ctx: Context) -> "Template":
        """Convert an Element to a Template object. Called from NTA.from_element."""
        template_obj = cls(ctx)

        template_obj.name = Name.from_element(et.find("name"))
        template_obj.parameter = Parameter.from_element(et.find("parameter"))
        template_obj.declaration = Declaration.from_element(et.find("declaration"))

        t_name = template_obj.name.name

        template_obj.graph = g.TAGraph(template_obj)
        template_obj.graph.template_name = t_name

        for l in et.iter("location"):
            loc = n.Location.from_element(l, ctx)
            template_obj.graph.add_location(loc)

        for b in et.iter("branchpoint"):
            bp = n.BranchPoint.from_element(b, ctx)
            template_obj.graph.add_branchpoint(bp)

        template_obj.graph.initial_location = et.find("init").get("ref")

        for t in et.iter("transition"):
            trans = tr.Transition.from_element(t, ctx)
            template_obj.graph.add_transition(trans)

        return template_obj

    def to_element(self):
        """Convert this object to an Element. Called from NTA.to_element."""
        element = ET.Element("template")
        element.append(self.name.to_element())
        if self.parameter is not None:
            element.append(self.parameter.to_element())
        if self.declaration is not None:
            element.append(self.declaration.to_element())
        element.extend(self.graph.to_element())
        return element



"""Unit tests for Locations and BranchPoints."""
import lxml.etree as ET
import pytest
from uppaalpy.classes.class_tests.test_context_cases import DataContext

from uppaalpy.classes.simplethings import ConstraintLabel
from uppaalpy.classes.transitions import Nail, Transition


class TestTransition:
    """Transition tests."""

    class C(DataContext):
        pass

    def test_transition_init(self):
        """Test Tranisiton()."""
        t1 = Transition(source="id1", target="id1")
        assert t1.source == "id1"
        assert t1.target == "id1"
        assert t1.nails == []

        t2 = Transition(source="id1", target="id2", nails=[Nail(1, 3), Nail(2, 2)])
        assert t2.nails[0].pos == (1, 3)
        assert t2.nails[1].pos == (2, 2)

        t3 = Transition(
            source="id1", target="id2", guard=ConstraintLabel("guard", "x == 0", (1, 3), self.C.ctx())
        )
        assert t3.guard is not None

        with pytest.raises(KeyError):
            assert Transition()

        with pytest.raises(KeyError):
            assert Transition(source="id1")

        with pytest.raises(KeyError):
            assert Transition(source="id2")

    def test_transition_from_element(self):
        """Test Transition.from_element()."""
        t1 = Transition.from_element(
            ET.fromstring(
                """
            <transition>
                    <source ref="id5"/>
                    <target ref="id17"/>
            </transition>
            """
            ), self.C.ctx()
        )

        assert t1.source == "id5"
        assert t1.target == "id17"
        assert t1.guard is None
        assert t1.assignment is None

        t2 = Transition.from_element(
            ET.fromstring(
                """
		<transition>
			<source ref="id5"/>
			<target ref="id17"/>
			<label kind="guard" x="-416" y="512">x == 10</label>
			<label kind="assignment" x="-416" y="528">counter=-1, detected=-1, slot_no[id]=-1,
aux_vec=zero_vec, first[id]=zero_vec, 
sent_info=0</label>
			<label kind="comments" x="-236" y="838">Broadcast information about collisions if any</label>
			<nail x="-72" y="544"/>
			<nail x="-464" y="544"/>
			<nail x="-464" y="40"/>
		</transition>
                """
            ), self.C.ctx()
        )

        assert t2.nails[2].pos == (-464, 40)
        assert t2.guard is not None
        assert t2.assignment is not None
        assert t2.comments is not None
        assert t2.synchronisation is None

    def test_transition_to_element(self):
        """Test Transition.to_element()."""
        e1 = Transition.from_element(
            ET.fromstring(
                """
            <transition>
                    <source ref="id5"/>
                    <target ref="id17"/>
            </transition>
            """
            ), self.C.ctx()
        ).to_element()

        assert e1.find("source").get("ref") == "id5"
        assert e1.find("target").get("ref") == "id17"

        e2 = Transition.from_element(
            ET.fromstring(
                """
		<transition>
			<source ref="id5"/>
			<target ref="id17"/>
			<label kind="guard" x="-416" y="512">x == 10</label>
			<label kind="assignment" x="-416" y="528">counter=-1, detected=-1, slot_no[id]=-1,
aux_vec=zero_vec, first[id]=zero_vec, 
sent_info=0</label>
			<label kind="comments" x="-236" y="838">Broadcast information about collisions if any</label>
			<nail x="-72" y="544"/>
			<nail x="-464" y="544"/>
			<nail x="-464" y="40"/>
		</transition>
                """
            ), self.C.ctx()
        ).to_element()

        labels = e2.findall("label")
        assert labels[0].get("kind") == "guard"
        assert labels[1].get("kind") == "assignment"
        assert labels[2].get("kind") == "comments"


class TestNail:
    """Nail tests."""

    @staticmethod
    def test_nail_init():
        """Test Nail()."""
        n = Nail(13, 26)
        assert n.pos == (13, 26)

        with pytest.raises(TypeError):
            assert Nail()  # Ignore the type error here by pylint/pyright.

    @staticmethod
    def test_nail_from_element():
        """Test Nail.from_element()."""
        n = Nail.from_element(ET.fromstring('<nail x="-72" y="544"/>'))
        assert n.pos == (-72, 544)

    @staticmethod
    def test_nail_to_element():
        """Test Nail.to_element()."""
        n = Nail(-72, 44)
        assert n.pos == (-72, 44)



"""Functions for path analysis."""
from collections import deque
from typing import Dict, List, Optional, Set, Tuple, Union, cast
from ortools.linear_solver import pywraplp
from itertools import product
from uppaalpy.classes.context import Context

from uppaalpy.classes.expr import ClockConstraintExpression
from uppaalpy.classes.nodes import Location
from uppaalpy.classes.templates import Template
from uppaalpy.classes.transitions import Transition


EPS = 2 ** (-10)

# _l : Location | Transition -> Location
# _t : Location | Transition -> Transition
_l = lambda x: cast(Location, x)
_t = lambda x: cast(Transition, x)

Path = List[Union[Location, Transition]]
LI = Tuple[str, str]  # ("id0", "template_name") : location identifier


def path_exists(path: Path) -> bool:
    """Return true if each edge is from previous location to next location."""
    for i in range(1, len(path) - 1, 2):
        if (
            _l(path[i - 1]).id != _t(path[i]).source
            or _l(path[i + 1]).id != _t(path[i]).target
        ):
            return False
    return True


def find_used_clocks(path: Path) -> List[str]:
    """Return the list of clocks that will be checked in path."""
    res = []
    for element in path:
        constraints = element.get_constraints()
        for c in constraints:
            if not isinstance(c, ClockConstraintExpression):
                continue
            for clock in c.clocks:
                if clock not in res:
                    res.append(clock)
    res.sort()
    return res


def convert_to_path(
    template: Template, lst: List[Union[int, str]]
) -> List[Union[Location, Transition]]:
    """Given a template and a list return a list of node and edges.

    Args:
        template: Template object.
        lst: Alternating list of location name strings and edge id's.
    Returns:
        An alternating list of location and transition objects.

    Example: extract_path(temp, ["l0", 1, "l2", 3, "l2"])
    returns an alternating list of location and edge objects.
    Locations are denoted with strings denoting their name fields.
    Edges are enumerated with zero-indexed integers according to their
    order in the xml file.
    This function makes use of nta.locations dictionary, hence it is assumed
    that each location in the graph is uniquely identified by their name field.
    """
    path = []
    for i, rep in enumerate(lst):
        if i % 2:
            path.append(template.graph._transitions[cast(int, rep)])
        else:
            path.append(template.graph._named_locations[cast(str, rep)])
    return path


def path_realizable(
    path: Path,
    validate_path: bool = False,
    validate_state: bool = False,
    add_epsilon: bool = False,
) -> Tuple[bool, List[float]]:
    """Given a path, construct an LP and return results.

    See: path_realizable_with_initial_valuation from this module.
    """
    return path_realizable_with_initial_valuation(
        path, validate_path, validate_state, add_epsilon
    )


def path_realizable_with_initial_valuation(
    path: Path,
    validate_path: bool = False,
    validate_state: bool = False,
    add_epsilon: bool = False,
    icv_constants: Optional[Dict[str, int]] = None,
    clocks: List[str] = None,
) -> Tuple[bool, List[float]]:
    """Given a path, construct an LP and return results.

    Args:
        path: List of alternating location and transitions.
        validate_path: Check whether each transition is actually between
            the previous location and the next location.
        validate_state: Whether to check non-clock constraints for path
            realizability.
        add_epsilon: Add epsilon to constraints with operators '<' and '>'.
            This is useful for invalidating solutions that would be otherwise
            correct with operators '<=' or '>='.
        icv_constants: Dict for the initial valuation of the clocks at the first
            location of the path.  If left as None, all initial clock valuations
            will be assumed to be 0.  Otherwise clock valuations with provided
            clock names as keys are initially the value corresponding to the clock
            name. Clock valuations omitted will be constrained with "c >= 0".
        clocks: List of clock name strings. If None, find_used_clocks function will be
            used to determine the used clocks by iterating the path.
    Returns:
        A tuple of a bool and a witness list of delays for each location.
    """
    # Strategy:
    #
    # If icv is None:
    #       variable count <- length(path)
    #       construct lp  -- each var corresponds to time spent on a location
    #
    # Otherwise:
    #       variable count <- clock count + length(path)
    #       construct lp with:
    #           Consider a clock val variable for each clock.
    #           Prepend each clock_to_delay[clock] with a corresponding icv var.
    #           Forall clocks: if initial clock value is specified:
    #               Constrain clock.

    # Check whether the path exists (not necessarily realizable).
    if validate_path and not path_exists(path):
        return False, []

    # We will construct the LP using these.
    # Each row in A correspond to a sequence of coefficients
    # ai1, ai2, ai3... ain such that ai1 * var1 + ai2 * var2 ... <= bi.
    # Matrix A and vector B will be populated by the compute_constraint
    # procedure.
    A = []
    B = []

    ctx = path[0].template.context.to_MutableContext()

    # Find clock names.
    if clocks == None:
        clocks = find_used_clocks(path)

    clock_to_delay: Dict[str, List[int]] = dict()

    delay_var_count = len(path) // 2
    var_count = delay_var_count  # var_count can be updated depending on icv.
    delay_var_offset = 0

    # Are icv (initial clock valuations) >= 0, == 0, or == some other value?

    if icv_constants is not None:
        # Initial clock valuations are not 0. Add a icv var to the LP for each
        # clock. For the function parameter icv, if icv_constants[clock_name] exists,
        # constrain that icv with icv <= icv_constants[clock_name]
        # and -icv <= -icv_constants[clock_name] (icv == icv_constants[clock_name].

        icv_var_count = len(clocks)
        var_count = delay_var_count + icv_var_count
        delay_var_offset = icv_var_count

        for i, c in enumerate(clocks):
            a = [[0] * var_count]
            b = [0]
            a[0][i] = -1  # Since (-var in -inf, 0) <=> (var in 0, inf)
            A.append(a[0])
            B.append(b[0])
            try:
                clock_to_delay[c] = [i, delay_var_offset]
                icv_c = icv_constants[c]
                b[0] = icv_c  # c in -var <= -c
                b.append(-icv_c)  # c in var <= c
                a.append([-var for var in a[0]])
                A.append(a[1])
                B.append(b[1])
            except:
                pass

        for i, c in enumerate(clocks):
            clock_to_delay[c] = [i, delay_var_offset]

    else:
        # Initial clock valuations are all 0. No need to consider icv vars.
        # delay_var_offset is 0.
        for c in clocks:
            clock_to_delay[c] = [0]

    # Generate LP coefficients for each constraint.
    for i in range(0, len(path) - 1, 2):
        # Source location
        l = _l(path[i])
        if l.invariant is not None:
            for c in l.invariant.constraints:
                if not isinstance(c, ClockConstraintExpression):
                    continue
                a, b = compute_constraint(
                    clock_to_delay, c, var_count, ctx, add_epsilon, validate_state
                )
                for k in range(len(a)):
                    A.append(a[k])
                    B.append(b[k])

        # Transition
        t = _t(path[i + 1])
        if t.guard is not None:
            for c in t.guard.constraints:
                if not isinstance(c, ClockConstraintExpression):
                    continue
                a, b = compute_constraint(
                    clock_to_delay, c, var_count, ctx, add_epsilon, validate_state
                )
                for k in range(len(a)):
                    A.append(a[k])
                    B.append(b[k])

        # Resets and updates
        resets_in_transition, updates_in_transition = [], []
        if t.assignment is not None:
            resets_in_transition = t.assignment.get_resets()
            updates_in_transition = t.assignment.get_other_updates()
        for c in resets_in_transition:
            clock_to_delay[c] = []
        for u in updates_in_transition:
            u.handle_update(ctx)

        # Target location
        l = _l(path[i + 2])
        if l.invariant is not None:
            for c in l.invariant.constraints:
                if not isinstance(c, ClockConstraintExpression):
                    continue
                a, b = compute_constraint(
                    clock_to_delay, c, var_count, ctx, add_epsilon, validate_state
                )
                for k in range(len(a)):
                    A.append(a[k])
                    B.append(b[k])

        # Add delays, consider delay_var_offset.
        for c in clocks:
            clock_to_delay[c].append(i // 2 + 1 + delay_var_offset)

    solver = pywraplp.Solver("", pywraplp.Solver.CBC_MIXED_INTEGER_PROGRAMMING)

    c = {}

    for j in range(var_count):
        c[j] = solver.NumVar(0, solver.infinity(), "x[%s]" % j)

    for i in range(len(A)):
        constraint = solver.RowConstraint(-solver.infinity(), B[i], "")
        for j in range(var_count):
            constraint.SetCoefficient(c[j], A[i][j])

    status = solver.Solve()

    delays = []

    if status == solver.OPTIMAL:
        for i in range(var_count):
            delays.append(c[i].solution_value())
        return True, delays

    if status == solver.INFEASIBLE:
        return False, []

    return False, []


def get_resets(transition, clocks):
    """Find clocks to be reset upon taking the transition.

    Args:
        transition: A transition object.
        clocks: List of used clocks in the analysis.

    Returns:
        list of clocks to be reset.
    """
    # TODO: delete me
    if transition.assignment is None:
        return []
    a_strings = [s.strip() for s in transition.assignment.value.split(",")]
    resets = []
    for assignment in a_strings:
        tokens = [word.strip().rstrip() for word in assignment.split("=")]
        if len(tokens) == 2 and tokens[1] == "0" and tokens[0] in clocks:
            resets.append(tokens[0])
    return resets


def compute_constraint(
    clock_to_delay: Dict[str, List[int]],
    exp: ClockConstraintExpression,
    variable_count: int,
    ctx: Context,
    add_epsilon: bool = False,
    skip_var_threshold: bool = False,
) -> Tuple[List[List[int]], List[int]]:
    """Construct a row of the Linear Program.

    The satisfiability of the constraints with variable thresholds depends on
    the current state. For this reason they should be omitted in
    semi-realizability checking for path segments considered while populating
    the DP table.

    Args:
        clock_to_delay: Dictionary of list of ints. clock_to_delay['x'] gives
            a list of delay variables for the locations visited since the last
            reset of clock 'x'.
        exp: ClockConstraintExpression
        variable_count: Int for determining the row size.
        ctx: Context used to determine the valuations of variables.
        add_epsilon: Whether to treat '<=' as '<' by adding an epsilon value to
            the threshold.
        skip_var_threshold: Whether to skip constraints of form clock < variable
    """

    if ctx.is_variable(exp.threshold) and skip_var_threshold:
        return [], []

    A_row = [[0 for _ in range(variable_count)]]
    for delay_var in clock_to_delay[exp.clocks[0]]:
        A_row[0][delay_var] = 1

    if len(exp.clocks) == 2:  # clock difference
        for delay_var in clock_to_delay[exp.clocks[1]]:
            A_row[0][delay_var] -= 1

    # TODO: get value from context
    B_row = [ctx.get_val(exp.threshold)]
    if exp.operator == ">":
        A_row[0] = [x * -1 for x in A_row[0]]
        B_row[0] = -1 * B_row[0]

    if add_epsilon and exp.equality == False:  # Inequality: Add epsilon.
        B_row[0] -= EPS

    if exp.operator == "=":
        A_row.append([x * -1 for x in A_row[0]])
        B_row.append(-B_row[0])

    return A_row, B_row


def find_all_semi_realizable_paths(
    template: Template, max_length: int
) -> Dict[LI, Dict[LI, List[List[Path]]]]:
    """Construct a DP table for all semi-realizable paths in the TA.

    A path is semi-realizable if and only if there exists initial clock
    valuations >= 0 such that that the path is realizable. A DP table is
    returned such that DP[i][j] is a collection (explained below) of all
    semi-realizable paths from the location i to location j.
    We assume there always exists a path from location x to location x of length 0
    (degenerate case).

    Args:
        template: A Template object.
        max_length: Check paths of length <= max_length. Assumed to be >= 1.

    Returns:
        A dict of dict of list of list of lists. DP['i']['j'][k] gives the
        list of semi-realizable paths from location named 'i' to location named
        'j' of length k.
    """
    g = template.graph
    nodes = g.nodes

    DP: Dict[LI, Dict[LI, List[List[Path]]]] = {}

    # Create DP table.
    for i, i_obj in nodes.data("obj"):  # type: ignore
        DP[i] = {}
        for j in nodes:
            DP[i][j] = [[] for _ in range(max_length + 1)]
        # Paths of length 0.
        DP[i][i][0].append([i_obj])

    # Add paths of length 1.
    for i, i_obj in nodes.data("obj"):  # type: ignore
        for j, edge_dict in g[i].items():
            j_obj = g.nodes.data("obj")[j]
            for e_attr in edge_dict.values():
                e_obj = e_attr["obj"]
                path = [i_obj, e_obj, j_obj]
                if path_realizable_with_initial_valuation(path, icv_constants={}):
                    DP[i][j][1].append(path)

    for l in range(2, max_length + 1):
        # Find paths of length l by examining subpaths of length (p) and (l - p).
        # p1 = i...j, p2 = j...k => p3 = i...j...k
        for p in range(1, l):
            s = l - p
            for i, j, k in product(nodes, repeat=3):
                for p1 in DP[i][j][p]:
                    for p2 in DP[j][k][s]:
                        p3 = p1[:-1] + p2
                        # c1 += l
                        # c2 += 1
                        if p3 in DP[i][k][l]:
                            continue
                        elif path_realizable_with_initial_valuation(
                            p3, icv_constants={}
                        ):
                            DP[i][k][l].append(p3)
    return DP


def check_dp(init: LI, target: LI, table: Dict[LI, Dict[LI, List[List[Path]]]]) -> bool:
    pps = table[init][target]
    for ps in pps:
        for p in ps:
            if path_realizable(p, True, True):
                return True
    return False


def find_reachable_locations(
    template: Template, table: Dict[LI, Dict[LI, List[List[Path]]]]
) -> None:
    """Mark the TAGraph of the given template for reachability.

    Reachability of a location can be checked with
    template.graph.nodes[("template_name", "location_id")]["tag"]
    """
    g = template.graph
    tn = g.template_name

    def f(x: str):
        return (tn, x)

    init = f(g.initial_location)

    for node in g:
        g.nodes[node]["tag"] = False
    q = deque()  # type: deque[LI]
    q.append(init)

    while q:
        n = q.popleft()
        if check_dp(init, n, table):
            g.nodes[n]["tag"] = True
            for succ in g.nodes[n]:
                q.append(succ)


def reachability_analysis(
    template: Template, table: Dict[LI, Dict[LI, List[List[Path]]]]
) -> Set[LI]:
    """Given a TA template and a path length limit, return reachable locations."""

    res = set()

    find_reachable_locations(template, table)

    for node, is_tagged in template.graph.nodes.data("tagged"):  # type: ignore
        if is_tagged:
            res.add(node)

    return res


def furthest_reachable(
    template: Template, target_set: Set[LI], table: Dict[LI, Dict[LI, List[List[Path]]]]
) -> Set[LI]:

    find_reachable_locations(template, table)
    gp = template.graph.reverse()
    q_act, q_alt = deque(target_set), deque()
    flag = False
    res: Set[LI] = set()

    while q_act:
        while q_act:
            n = q_act.popleft()
            if gp.nodes[n]["tag"]:
                flag = True
                res.add(n)
            if not flag:
                for succ in gp[n]:
                    q_alt.append(succ)
        if not flag:
            # breaks from the outer loop if q_alt is empty or flag is True
            q_act = q_alt
            q_alt = deque()

    return res



def concatenate_paths(template, p1, p2):
    """Return a list of possible concatenations of given two paths.

    A distinct path for each transition going from the last location of p1 to
    the initial location of p2 is created.

    Args:
        template: A NTA object.
        p1: A list denoting a path. l1-t1-l2-t2..ln
        p2: A list denoting a path. l(n+1)-tn..lm

    Returns:
        A list of paths l1..ln-tx-l(n+1)..lm for each tx.
    """
    # TODO: delete me
    res = []
    last = p1[-1].id
    first = p2[0].id
    g = template.graph
    t_name = g.template_name

    # If no transition between the two exists, key error exception is thrown.
    try:
        # Iterate over all transitions from last to first.
        for obj in g[(t_name, last)][(t_name, first)].values():
            t = obj["obj"]
            p = []
            p.extend(p1)
            p.append(t)
            p.extend(p2)
            res.append(p)
    except KeyError:
        pass

    return res


if __name__ == "__main__":
    import uppaalpy as u

    temp = u.NTA.from_xml("examples/generator/test5_6_2.xml").templates[0]
    mypath = convert_to_path(temp, ["l0", 0, "l2", 1, "l3", 2, "l4"])



"""Tests for the ConstraintCache and ConstraintPatch classes.

A complete NTA object is required to test these classes.
"""

from uppaalpy.classes.constraint_patcher import ConstraintPatch, ConstraintUpdate
from uppaalpy.classes.nta import NTA

from .helpers import testcase_dir


class TestConstraintPatch:
    """Unit test for ConstraintPatch."""

    @staticmethod
    def test_constraint_patch_init_transition():
        """Test ConstraintPatch()."""
        nta = NTA.from_xml(testcase_dir + "constraint_cache_xml_files/test01.xml")

        # Update the threshold of the transition in the second template.
        # x <= 10 -> x <= 15
        template = nta.templates[1]  # Second template
        transition = template.graph._transitions[0]  # First transition
        constraints = transition.guard.constraints

        update = ConstraintUpdate(constraints[0], 15)  # type: ignore

        cp = ConstraintPatch(template, update, obj_ref=transition)

        assert cp.template_ref == template
        assert cp.obj_ref == transition
        assert cp.change == update

    @staticmethod
    def test_constraint_patch_init_location():
        """Test ConstraintPatch()."""
        nta = NTA.from_xml(testcase_dir + "constraint_cache_xml_files/test01.xml")

        # Update the invariant of l0 in the first template.
        # x <= 10 -> x <= 15
        template = nta.templates[0]  # First template
        location = template.graph._named_locations["l0"]
        constraints = location.invariant.constraints

        update = ConstraintUpdate(constraints[0], 15)  # type: ignore

        cp = ConstraintPatch(template, update, obj_ref=location)

        assert cp.template_ref == template
        assert cp.obj_ref == location
        assert cp.change == update


class TestConstraintCache:
    """Unit tests for ConstraintCache."""

    @staticmethod
    def test_constraint_cache_init():
        """Test ConstraintCache()."""
        nta = NTA.from_xml(testcase_dir + "constraint_cache_xml_files/test01.xml")
        cc = nta.patch_cache

        assert cc.nta == nta
        assert cc.patches == []

    @staticmethod
    def test_constraint_cache_cache():
        """Test ConstraintCache.cache()."""
        nta = NTA.from_xml(testcase_dir + "constraint_cache_xml_files/test01.xml")
        cc = nta.patch_cache

        # Update the threshold of the transition in the second template.
        # x <= 10 -> x <= 15
        template = nta.templates[1]  # Second template
        transition = template.graph._transitions[0]  # First template
        constraints = transition.guard.constraints

        update = ConstraintUpdate(constraints[0], 15)  # type: ignore

        cp = ConstraintPatch(template, update, obj_ref=transition)
        cc.cache(cp)

        assert cc.patches == [cp]

    @staticmethod
    def test_constraint_cache_apply_single_patch():
        """Test _apply_single_patch method."""
        nta = NTA.from_xml(testcase_dir + "constraint_cache_xml_files/test01.xml")
        cc = nta.patch_cache

        # Update the threshold of the transition in the second template.
        # x <= 10 -> x <= 15
        template = nta.templates[1]  # Second template
        transition = template.graph._transitions[0]  # First template
        constraints = transition.guard.constraints

        update = ConstraintUpdate(constraints[0], "15") # type: ignore

        cp = ConstraintPatch(template, update, obj_ref=transition)
        cc.cache(cp)

        lines = open(testcase_dir + "constraint_cache_xml_files/test01.xml").readlines()
        lines_subject_to_change = lines[:]

        cc._apply_single_patch(lines_subject_to_change, cc.patches[0])

        # 56th line is changed.
        assert lines[:56] == lines_subject_to_change[:56]
        assert lines[57:] == lines_subject_to_change[57:]
        assert lines[56].replace("10", "15") == lines_subject_to_change[56]



"""Subclass of networkx multidigraph."""

from itertools import count
from typing import Dict, List

import lxml.etree as ET
import networkx as nx

from uppaalpy.classes import nodes as n
from uppaalpy.classes import templates as te
from uppaalpy.classes import transitions as tr


class TAGraph(nx.MultiDiGraph):
    """Derived class of NetworkX MultiDiGraph.

    Do not use typical networkx methods to modify the graph.
    Extends the base class with following attributes.

    Extra attributes:
        initial_location: String for initial location ref. (ex: "id0")
        named_locations: Dictionary for mapping from location names to locations.
        transitions: List of transitions, in the order read from file.
        template_name: String for name of the template.
        transition_counter: Iterator for uniquely assigning a uniquely determined
            key to each transition in the MultiDiGraph.
        template: The parent template.
    """

    def __init__(self, template, incoming_graph_data=None, **attr):
        """Create a TAGraph.

        Superclass initializer is called. Template name and object are set
        by Template.from_element(). Other attributes are updated/set by
        methods add_{location,branchpoint,transition}.
        """
        super().__init__(incoming_graph_data, **attr)
        self.initial_location = ""  # type: str
        self._named_locations = {}  # type: Dict[str, n.Location]
        self._transitions = []  # type: List[tr.Transition]
        self.template_name = ""  # type: str
        self._transition_counter = count()
        self.template = template  # type: te.Template

    def add_location(self, loc):
        """Insert a Location object.

        Only named Locations can be used for path analysis. Named Locations
        are also registered in self._named_locations.
        """
        loc.template = self.template
        self.add_node((self.template_name, loc.id), obj=loc)
        if loc.name != None:
            self._named_locations[loc.name.name] = loc

    def add_branchpoint(self, bp):
        """Insert a BranchPoint object. See add_location()."""
        self.add_node((self.template_name, bp.id), obj=bp)

    def add_transition(self, trans):
        """Insert a Transition object.

        self._transition_counter is used for manually assigning a unique
        key to the edge. Also, self._transitions is used for linear time
        serializations and constant time lookups.
        """
        trans.template = self.template
        self.add_edge(
            (self.template_name, trans.source),
            (self.template_name, trans.target),
            obj=trans,
            key=next(self._transition_counter),
        )
        self._transitions.append(trans)

    def to_element(self):
        """Convert the multidigraph to an Element."""
        elements = [n.to_element() for n in self.get_nodes()]
        elements.append(ET.Element("init", attrib={"ref": self.initial_location}))
        elements.extend([t.to_element() for t in self._transitions])
        return elements

    def get_nodes(self):
        """Get the list of nodes. Also includes branchpoints."""
        return [data["obj"] for _, data in list(self.nodes(data=True))]



"""Class definition of UPPAAL NTA."""
from typing import Callable, List, Optional, Type, Union, cast

import lxml.etree as ET

from uppaalpy.classes import constraint_patcher as cp
from uppaalpy.classes import nodes as n
from uppaalpy.classes import templates as te
from uppaalpy.classes import transitions as tr
from uppaalpy.classes.context import Context
from uppaalpy.classes.expr import ClockConstraintExpression
from uppaalpy.classes.simplethings import Declaration, Query, SystemDeclaration


class NTA:
    """Class that holds information about a network of timed automata.

    Attributes:
        declaration: Declaration object.
        context: Context object generated by parsing the declaration.
        templates: List of template objects.
        system: SystemDeclaration object.
        queries: List of Query objects.
        patch_cache: Constraint cache object. Changes on constraints are stored
            here for fast write operations.
        _associated_file: String for denoting the path of the file read. Used
            by constraint patcher. Set by the from_xml method.
        _doctype: String for xml doctype. Set by from_xml, and used by to_file.
    """

    def __init__(self, **kwargs) -> None:
        """Initialize NTA from keyword arguments.

        Args:
            **kwargs: Keyword arguments for initializing the NTA. See class docstring.
        """
        self.declaration = kwargs.get("declaration")  # type: Optional[Declaration]
        self.context = kwargs["context"]  # type: Context
        self.templates = kwargs.get("templates") or []  # type: List[te.Template]
        self.system = kwargs["system"]  # type: SystemDeclaration
        self.queries = kwargs["queries"]  # type: List[Query]
        self.patch_cache = cp.ConstraintCache(self)
        self._associated_file: str = ""
        self._doctype: str = ""

    @classmethod
    def from_xml(cls: Type["NTA"], path: str) -> "NTA":
        """Given a xml file path, construct an NTA from that xml file."""
        et = ET.parse(path)
        obj = cls.from_element(et.getroot())
        obj._associated_file = path
        obj._doctype = et.docinfo.doctype
        return obj

    @classmethod
    def from_element(cls: Type["NTA"], et) -> "NTA":
        """Construct NTA from Element, and return it."""
        kw = {}
        kw["declaration"] = Declaration.from_element(et.find("declaration"))

        if kw["declaration"] is None or kw["declaration"].text is None:
            kw["context"] = Context.parse_context(None)
        else:
            kw["context"] = Context.parse_context(kw["declaration"].text)
        kw["templates"] = [
            te.Template.from_element(template, kw["context"])
            for template in et.iterchildren("template")
        ]
        kw["system"] = SystemDeclaration.from_element(et.find("system"))
        if et.find("queries") is None:
            kw["queries"] = []
        else:
            kw["queries"] = [
                Query.from_element(query) for query in et.find("queries").iter("query")
            ]
        return cls(**kw)

    def to_element(self):
        """Construct an Element object, and return it."""
        root = ET.Element("nta")

        if self.declaration is not None:
            et_declaration = ET.SubElement(root, "declaration")
            et_declaration.text = self.declaration.text

        for template in self.templates:
            root.append(template.to_element())

        root.append(self.system.to_element())
        queries = ET.SubElement(root, "queries")
        for query in self.queries:
            queries.append(query.to_element())

        return root

    def to_file(self, path: str) -> None:
        r"""Convert the NTA to an element tree and write it into a file.

        File will be printed with pretty printing and '\t' indentation.

        Args:
            path: String denoting the path of the output file.
        """
        elm = self.to_element()
        ET.indent(elm, "\t")
        ET.ElementTree(elm).write(
            path, encoding="utf-8", xml_declaration=True, doctype=self._doctype
        )

    def change_clock_constraint(
        self,
        obj: Union[tr.Transition, n.Location],
        operation: str,
        ccexp: ClockConstraintExpression,
        *,
        threshold_function: Optional[Callable[[str], str]] = None,
    ) -> None:
        """Insert/remove/update a simple constraint expression.

        A constraint patch is created to be cached in the patch cache for the
        changed constraint. Operations can be "insert", "remove", or "update".

        Args:
            obj: Either a Transition or a Location.
            operation: "insert", "remove", or "update".
            ccexp: ClockConstraintExpression object.
                Insert mode: ClockConstraintExpression to be inserted to the
                    guard/invariant label.
                Remove mode: ClockConstraintExpression to be removed from the
                    guard/invariant label.
                Update mode: ClockConstraintExpression whose threshold will be
                    changed.
            threshold_function: Function f: str -> str used in Update mode. The
                treshold x is set to threshold_function(x). Notice that x is
                either an identifier or a value. In both cases its type is str.

        Insert mode:
        A new ClockConstraintExpression is inserted to the guard/invariant
        label. If no guard exists, one is created located between the source and
        target for guards, and on the location for invariants. A
        ConstraintInsert is created and cached in the patch cache.

        Remove mode:
        A ClockConstraintExpression is removed from the guard/invariant. If the
        expression to be removed is the only remaining constraint on the object,
        the label is removed, as well. A ConstraintRemove is created and cached.

        Update mode:
        The ClockConstraintExpression object is updated with the threshold_function.
        A ConstraintUpdate is created and cached.
        """
        label = obj.get_constraint_label()

        if operation == "insert":
            if label is None:  # Create guard or invariant
                obj.create_constraint_label(ccexp, self.context)
                change = cp.ConstraintInsert(
                    ccexp, newly_created=obj.get_constraint_label()
                )
            else:
                label.constraints.append(ccexp)
                change = cp.ConstraintInsert(ccexp)

        elif operation == "remove":
            label.constraints.remove(ccexp)
            if label.constraints == []:
                change = cp.ConstraintRemove(ccexp, True)
                obj.remove_constraint_label()
            else:
                change = cp.ConstraintRemove(ccexp)

        else:  # operation == "update"
            new = threshold_function(ccexp.threshold)
            change = cp.ConstraintUpdate(ccexp, new)
            index = label.constraints.index(ccexp)
            label.constraints.pop(index)
            label.constraints.insert(index, change.generate_new_constraint())

        patch = cp.ConstraintPatch(cast(te.Template, obj.template), change, obj)
        self.patch_cache.cache(patch)

    def flush_constraint_changes(self, out_path):
        """Read the associated_file and write the modified version to new file.

        The xml file the NTA is constructed from is read again and the lines
        are updated according to the patch_cache's update records.
        See: constraint_patcher.py
        """
        with open(self._associated_file) as input_file:
            lines = input_file.readlines()

        self.patch_cache.apply_patches(lines)

        with open(out_path, "w") as output_file:
            output_file.writelines(lines)



"""Unit tests for Locations and BranchPoints."""
from typing import cast
import lxml.etree as ET
import pytest
from uppaalpy.classes.class_tests.test_context_cases import DataContext
from uppaalpy.classes.expr import ClockConstraintExpression

from uppaalpy.classes.nodes import BranchPoint, Location, Node
from uppaalpy.classes.simplethings import ConstraintLabel, Name


class TestNode:
    """Node tests."""
    class C(DataContext):
        pass

    def test_generate_dict(self):
        """Test initialization data generated from an Element."""
        n1 = Node.generate_dict(
            ET.fromstring(
                """
        <location id="id1" x="192" y="384">
            <name x="208" y="392">Stop</name>
        </location>
            """
            ), self.C.ctx()
        )
        assert n1["id"] == "id1"
        assert n1["pos"] == (192, 384)
        assert n1["name"].name == "Stop"
        assert n1["name"].pos == (208, 392)

        n2 = Node.generate_dict(
            ET.fromstring(
                """
        <location id="id3" x="96" y="256">
            <name x="40" y="240">Appr</name>
            <label kind="invariant" x="32" y="264">x&lt;=20</label>
        </location>
            """
            ), self.C.ctx()
        )
        assert n2["id"] == "id3"
        assert n2["pos"] == (96, 256)
        assert n2["name"].name == "Appr"
        assert n2["name"].pos == (40, 240)
        assert n2["invariant"].kind == "invariant"
        assert n2["invariant"].value == "x<=20"
        assert n2["invariant"].pos == (32, 264)

        n3 = Node.generate_dict(
            ET.fromstring(
                """
                <location id="id5" x="144" y="-88">
                        <name x="134" y="-118">C1</name>
                        <label kind="invariant" x="152" y="-72">x'==2</label>
                        <label kind="exponentialrate" x="134" y="-73">3</label>
                </location>
            """
            ), self.C.ctx()
        )
        assert n3["id"] == "id5"
        assert n3["pos"] == (144, -88)
        assert n3["name"].name == "C1"
        assert n3["name"].pos == (134, -118)
        assert n3["invariant"].kind == "invariant"
        assert n3["invariant"].value == "x'==2"
        assert n3["invariant"].pos == (152, -72)
        assert n3["exponentialrate"].kind == "exponentialrate"
        assert n3["exponentialrate"].value == "3"
        assert n3["exponentialrate"].pos == (134, -73)

        n4 = Node.generate_dict(
            ET.fromstring(
                """
                <location id="id2" x="8" y="-17">
                        <name x="-26" y="-17">Off</name>
                        <label kind="testcodeEnter">expect_off();</label>
                </location>
            """
            ), self.C.ctx()
        )
        assert n4["id"] == "id2"
        assert n4["pos"] == (8, -17)
        assert n4["name"].name == "Off"
        assert n4["name"].pos == (-26, -17)
        assert n4["testcodeEnter"].kind == "testcodeEnter"
        assert n4["testcodeEnter"].value == "expect_off();"
        assert n4["testcodeEnter"].pos == None
        assert n4["committed"] == False

        n5 = Node.generate_dict(
            ET.fromstring(
                """
                <location id="id9" x="88" y="336">
                        <committed/>
                </location>
            """
            ), self.C.ctx()
        )
        assert n5["id"] == "id9"
        assert n5["pos"] == (88, 336)
        assert n5["committed"] == True


class TestBranchPoint:
    """Unit tests for branchpoints."""
    class C(DataContext):
        pass

    @staticmethod
    def test_branchpoint_init():
        """Test the init method."""
        bp = BranchPoint(id="id3", pos=(3, 4))
        assert bp.id == "id3"
        assert bp.pos == (3, 4)
        assert bp.tag == "branchpoint"

        bp = BranchPoint(id="id15", pos=(-13, 7))
        assert bp.id == "id15"
        assert bp.pos == (-13, 7)
        assert bp.tag == "branchpoint"

        with pytest.raises(KeyError):
            assert BranchPoint(id=13)

    def test_branchpoint_from_element(self):
        """Test the from_element method."""
        xml1 = ET.fromstring('<branchpoint id="id4" x="-25" y="-25"></branchpoint>')
        bp1 = BranchPoint.from_element(xml1, self.C.ctx())

        assert bp1.id == "id4"
        assert bp1.pos == (-25, -25)

        xml2 = ET.fromstring('<branchpoint id="id7" x="48" y="-48"></branchpoint>')
        bp2 = BranchPoint.from_element(xml2, self.C.ctx())

        assert bp2.id == "id7"
        assert bp2.pos == (48, -48)

    def test_branchpoint_to_element(self):
        """Test the to_element method."""
        xml1 = ET.fromstring('<branchpoint id="id4" x="-25" y="-25"></branchpoint>')
        e1 = BranchPoint.from_element(xml1, self.C.ctx()).to_element()

        assert e1.get("id") == "id4"
        assert e1.get("x") == "-25"
        assert e1.get("y") == "-25"

        xml2 = ET.fromstring('<branchpoint id="id7" x="48" y="-48"></branchpoint>')
        e2 = BranchPoint.from_element(xml2, self.C.ctx()).to_element()

        assert e2.get("id") == "id7"
        assert e2.get("x") == "48"
        assert e2.get("y") == "-48"


class TestLocation:
    """Tests for Location class."""

    class C(DataContext):
        pass

    def test_location_init(self):
        """Test the init method."""
        l1 = Location(id="id1", pos=(1, 3))
        assert l1.id == "id1"
        assert l1.pos == (1, 3)

        l2 = Location(id="id1", pos=(1, 3), name=Name("loc1", (4, 8)))
        assert l2.id == "id1"
        assert l2.pos == (1, 3)
        assert l2.pos == (1, 3)
        assert l2.name.name == "loc1"
        assert l2.name.pos == (4, 8)

        l3 = Location(
            id="id1",
            pos=(1, 3),
            name=Name("loc1", (4, 8)),
            invariant=ConstraintLabel("invariant", "x <= 13", (14, 28), self.C.ctx()),
        )
        assert l3.id == "id1"
        assert l3.pos == (1, 3)
        assert l3.pos == (1, 3)
        assert l3.name.name == "loc1"
        assert l3.name.pos == (4, 8)
        assert l3.invariant.kind == "invariant"

        with pytest.raises(KeyError):
            assert Location(id="id1")

        with pytest.raises(KeyError):
            assert Location(pos=(1, 3))

    def test_location_from_element(self):
        """Test the from_element method."""
        l1 = Location.from_element(
            ET.fromstring(
                """
        <location id="id1" x="192" y="384">
            <name x="208" y="392">Stop</name>
        </location>
            """
            ), self.C.ctx()
        )

        assert l1.id == "id1"
        assert l1.pos == (192, 384)
        assert l1.committed == False
        assert l1.urgent == False

        l2 = Location.from_element(
            ET.fromstring(
                """
        <location id="id3" x="96" y="256">
            <name x="40" y="240">Appr</name>
            <label kind="invariant" x="32" y="264">x&lt;=20</label>
            <label kind="exponentialrate" x="134" y="-73">3</label>
            <label kind="testcodeEnter">expect_off();</label>
            <committed/>
        </location>
            """
            ), self.C.ctx()
        )

        assert l2.id == "id3"
        assert l2.pos == (96, 256)
        assert l2.committed == True
        assert l2.urgent == False
        assert l2.invariant is not None
        assert l2.testcodeEnter is not None

    def test_location_to_element(self):
        """Test the to_element method."""
        e1 = Location.from_element(
            ET.fromstring(
                """
        <location id="id1" x="192" y="384">
            <name x="208" y="392">Stop</name>
        </location>
            """
            ), self.C.ctx()
        ).to_element()

        assert e1.get("id") == "id1"
        assert e1.get("x") == "192"
        assert e1.get("y") == "384"

        e2 = Location.from_element(
            ET.fromstring(
                """
        <location id="id3" x="96" y="256">
            <name x="40" y="240">Appr</name>
            <label kind="invariant" x="32" y="264">x&lt;=20</label>
            <label kind="exponentialrate" x="134" y="-73">3</label>
            <label kind="testcodeEnter">expect_off();</label>
            <committed/>
        </location>
            """
            ), self.C.ctx()
        ).to_element()

        assert e2.get("id") == "id3"
        assert e2.get("x") == "96"
        assert e2.get("y") == "256"
        labels = e2.findall("label")
        assert labels[0].get("kind") == "invariant"
        assert labels[1].get("kind") == "exponentialrate"
        assert labels[2].get("kind") == "testcodeEnter"
        assert e2.find("committed") in e2
        assert e2.find("urgent") not in e2

    def test_location_get_constrasints(self):
        """Test Location.get_constraints()."""
        l1 = Location.from_element(
            ET.fromstring(
                """
        <location id="id1" x="192" y="384">
            <name x="208" y="392">Stop</name>
        </location>
            """
            ), self.C.ctx()
        )

        assert l1.get_constraints() == []

        l2 = Location.from_element(
            ET.fromstring(
                """
        <location id="id3" x="96" y="256">
            <name x="40" y="240">Appr</name>
            <label kind="invariant" x="32" y="264">c&lt;=20</label>
            <label kind="exponentialrate" x="134" y="-73">3</label>
            <label kind="testcodeEnter">expect_off();</label>
            <committed/>
        </location>
            """
            ), self.C.ctx()
        )

        cs = l2.get_constraints()
        assert len(cs) == 1
        c = cast(ClockConstraintExpression, cs[0])
        assert c.clocks == ["c"]
        assert c.operator == "<"
        assert c.threshold == "20"
        assert c.equality == True



from itertools import product

from lxml import etree as ET
from pytest_cases import parametrize
from uppaalpy.classes.class_tests.test_context_cases import DataContext

from uppaalpy.classes.expr import ClockResetExpression, UpdateExpression
from uppaalpy.classes.simplethings import UpdateLabel

# from uppaalpy.classes.class_tests.test_context_cases import DataContext


class CaseUpdateLabel:
    pass
    label_kinds = ["assignment"]
    # ctx = DataContext.ctx
    label_vals = [
        "c1 = 0",
        "c1 = 0, c2 = 0",
        "i += 1",
        "i = 1",
        "i = k, k = j, j += i",
        "c2 = 0, i = constant3",
        "k -= y",
    ]


class DataLabel:
    label_kinds = [
        "synchronisation",
        "assignment",
        "exponentialrate",
        "testcode",
        "testcodeEnter",
        "testcodeExit",
        "select",
    ]

    label_vals = [
        "foo",
        "bar",
        "x <= 10",
        "x >= 15",
    ]

    label_pos = [(0, 0), (1, 1), (10, 20), (-15, 28)]

    label_element_str = [
        """<label kind="invariant" x="-128" y="64">t&lt;=T[id] &amp;&amp; wcrt'==0</label>""",
        """<label kind="invariant" x="48" y="48">t&lt;=L[id] &amp;&amp; wcrt'==0</label>""",
        """<label kind="invariant" x="-16" y="336">ax&lt;=ci[id][r][0]</label>""",
        """<label kind="invariant" x="8" y="496">ax'==0</label>""",
        """<label kind="guard" x="-24" y="64">t==T[id]</label>""",
        """<label kind="synchronisation" x="-208" y="312">ready!</label>""",
        """<label kind="assignment" x="-208" y="328">updatePriority(ci[id][r][1]), r++</label>""",
        """<label kind="testcodeExit">updatePriority(ci[id][r][1]), r++</label>""",
    ]

    label_kv = product(label_kinds, label_vals)
    label_kvp = product(label_kinds, label_vals, label_pos)
    label_element = [ET.fromstring(s) for s in label_element_str]


class CaseLabelInit:
    class C(DataLabel):
        pass

    @parametrize("kind, val", zip(C.label_kinds, C.label_vals))
    def case_no_position(self, kind, val):
        return (kind, val, None)

    @parametrize("kind, val, pos", zip(C.label_kinds, C.label_vals, C.label_pos))
    def case_with_position(self, kind, val, pos):
        return (kind, val, pos)


class CaseLabelFromElement:
    class C(DataLabel):
        pass

    @parametrize("e", C.label_element)
    def case_1(self, e):
        return e


class CaseLabelToElement:
    class C(DataLabel):
        pass

    @parametrize("e", C.label_element)
    def case_1(self, e):
        return e

class CaseUpdateLabelInit:
    class C(DataLabel, DataContext):
        pass

    def case_updates_none(self):
        ctx = self.C.ctx()
        return "assignment", "i += 15", (0, 1), ctx, None

    def case_updates_not_none(self):
        ctx = self.C.ctx()
        e = UpdateExpression("i += 15", ctx)
        return "assignment", "i += 15", (0, 1), ctx, [e]

    def case_updates_multi(self):
        ctx = self.C.ctx()
        e1 = UpdateExpression("i += 15", ctx)
        e2 = ClockResetExpression("c = 0", ctx)
        return "assignment", "i += 15", (0, 1), ctx, [e1, e2]

class CaseUpdateResets:
    class C(DataLabel, DataContext):
        pass

    def case_01(self):
        return UpdateLabel("assignment", "i += 10", (0, 0), self.C.ctx()), []

    def case_02(self):
        return UpdateLabel("assignment", "c = 0", (0, 0), self.C.ctx()), ["c"]

    def case_03(self):
        return UpdateLabel("assignment", "c = 0, c1 = 0, i += 10", (0, 0), self.C.ctx()), ["c", "c1"]



"""Unit tests for Templates and TAGraphs."""
import lxml.etree as ET
from uppaalpy.classes.class_tests.test_context_cases import DataContext
from uppaalpy.classes.nodes import Location
from uppaalpy.classes.simplethings import ConstraintLabel, Name
from uppaalpy.classes.tagraph import TAGraph
from uppaalpy.classes.templates import Template
from uppaalpy.classes.transitions import Transition

from .helpers import testcase_dir


class TestTemplate:
    """Template tests."""
    class C(DataContext):
        pass

    def test_template_init(self):
        """Test default values set by Template()."""
        template = Template(self.C.ctx())
        assert template.name.name == ""
        assert template.name.pos == (0, 0)
        assert template.parameter.tag == "parameter"
        assert template.parameter.text == ""
        assert template.declaration.tag == "declaration"
        assert template.declaration.text == ""

    def test_template_from_element(self):
        """Test Template.from_element()."""
        t = Template.from_element(
            ET.parse(testcase_dir + "template_xml_files/test1.xml").getroot(),
            self.C.ctx()
        )
        assert t.name.name == "Test1"
        assert t.parameter == None
        assert t.declaration is not None

        t = Template.from_element(
            ET.parse(testcase_dir + "template_xml_files/test2.xml").getroot(),
            self.C.ctx()
        )
        assert t.name.name == "Test2"
        assert t.parameter == None
        assert t.declaration is not None

        t = Template.from_element(
            ET.parse(testcase_dir + "template_xml_files/test3.xml").getroot(),
            self.C.ctx()
        )
        assert t.name.name == "P"
        assert t.parameter.text == "const id_t pid"
        assert t.declaration.text == "clock x;\nconst int k = 2;"

        t = Template.from_element(
            ET.parse(testcase_dir + "template_xml_files/test4.xml").getroot(),
            self.C.ctx()
        )
        assert t.name.name == "Train"
        assert t.parameter.text == "const id_t id"
        assert t.declaration.text == "clock x;"

    def test_template_to_element(self):
        """Test Template.to_element()."""
        t = Template.from_element(
            ET.parse(testcase_dir + "template_xml_files/test1.xml").getroot(), self.C.ctx()
        ).to_element()
        assert t.find("name").text == "Test1"
        assert t.find("parameter") == None

        t = Template.from_element(
            ET.parse(testcase_dir + "template_xml_files/test2.xml").getroot(), self.C.ctx()
        ).to_element()
        assert t.find("name").text == "Test2"
        assert t.find("parameter") == None

        t = Template.from_element(
            ET.parse(testcase_dir + "template_xml_files/test3.xml").getroot(), self.C.ctx()
        ).to_element()
        assert t.find("name").text == "P"
        assert t.find("parameter").text == "const id_t pid"
        assert t.find("declaration").text == "clock x;\nconst int k = 2;"

        t = Template.from_element(
            ET.parse(testcase_dir + "template_xml_files/test4.xml").getroot(), self.C.ctx()
        ).to_element()
        assert t.find("name").text == "Train"
        assert t.find("parameter").text == "const id_t id"
        assert t.find("declaration").text == "clock x;"


class TestTAGraph:
    """Unit tests for TAGraphs."""

    class C(DataContext):
        pass

    @staticmethod
    def test_tagraph_init():
        """Test initial values set by TAGraph()."""
        graph = TAGraph(None)
        assert graph.initial_location == ""
        assert graph.template == None
        assert graph.template_name == ""
        assert graph._named_locations == {}
        assert graph._transitions == []

    def test_tagraph_add_location(self):
        """Test adding a location to the graph."""
        graph = TAGraph(None)
        loc = Location(
            id="id0",
            pos=(0, 0),
            name=Name("loc0", (0, 0)),
            invariant=ConstraintLabel("invariant", "x <= 3", (0, 0), self.C.ctx()),
        )
        graph.add_location(loc)
        assert graph._named_locations["loc0"] is not None
        assert graph.nodes(data="obj")[("", "id0")] == loc

        loc2 = Location(
            id="id1",
            pos=(1, 1),
            invariant=ConstraintLabel("invariant", "x <= 4", (1, 1), self.C.ctx()),
        )
        graph.add_location(loc2)
        assert len(graph._named_locations.keys()) == 1
        assert graph.nodes(data="obj")[("", "id1")] == loc2

    def test_tagraph_add_transition(self):
        """Test adding a transition to the graph."""
        graph = TAGraph(None)
        loc = Location(
            id="id0",
            pos=(0, 0),
            name=Name("loc0", (0, 0)),
            invariant=ConstraintLabel("invariant", "x <= 3", (0, 0), self.C.ctx()),
        )
        graph.add_location(loc)

        loc2 = Location(
            id="id1",
            pos=(1, 1),
            invariant=ConstraintLabel("invariant", "x <= 4", (1, 1), self.C.ctx()),
        )
        graph.add_location(loc2)

        trans = Transition(source="id0", target="id1")
        graph.add_transition(trans)

        assert graph._transitions[0] == trans
        assert graph[("", "id0")][("", "id1")][0]["obj"] == trans

    def test_tagraph_init_with_template(self):
        """Test TAGraph initialization."""
        t = Template.from_element(
            ET.parse(testcase_dir + "template_xml_files/test1.xml").getroot(), self.C.ctx()
        )
        g = t.graph
        assert g.template == t
        assert g.template_name == t.name.name
        assert list(g._named_locations.keys()) == ["l2", "l1", "l0"]
        assert g._named_locations["l2"].id == "id0"
        assert g._named_locations["l1"].id == "id1"
        assert g._named_locations["l0"].id == "id2"
        assert g.initial_location == "id2"

        assert g._transitions[0].source == "id2"
        assert g._transitions[1].source == "id1"
        assert g._transitions[2].source == "id1"
        assert g._transitions[3].source == "id2"

        assert g._transitions[0].target == "id0"
        assert g._transitions[1].target == "id0"
        assert g._transitions[2].target == "id0"
        assert g._transitions[3].target == "id1"

    def test_tagraph_get_nodes(self):
        """Test TAGraph.get_nodes()."""
        graph = TAGraph(None)
        loc = Location(
            id="id0",
            pos=(0, 0),
            name=Name("loc0", (0, 0)),
            invariant=ConstraintLabel("invariant", "x <= 3", (0, 0), self.C.ctx()),
        )
        graph.add_location(loc)

        loc2 = Location(
            id="id1",
            pos=(1, 1),
            invariant=ConstraintLabel("invariant", "x <= 4", (1, 1), self.C.ctx()),
        )
        graph.add_location(loc2)

        nlist = graph.get_nodes()
        assert nlist[0] == loc
        assert nlist[1] == loc2



"""Classes for representing various expressions."""

from abc import ABCMeta, abstractmethod
from typing import List, Literal, Sequence, Tuple, TypeVar, Union

from uppaalpy.classes import context as c # import Context, MutableContext

E = TypeVar("E", bound="Expression")


class Expression(metaclass=ABCMeta):
    """Abstact base class representing expressions.

    An expression can either be an update to the current variables or
    evaluate to a boolean value. Currently, not all expressions in UPPAAL
    are allowed. For details, see child classes ConstraintExpression and
    UpdateExpression.

    Attributes:
        delimiter: Delimiter used for splitting/joining complex expressions.
        lhs: Left hand side of the expression.
        op: Operator
        rhs: Right hand side of the expression

    Methods:
        get_delimiter: Delimiter used for splitting/joining complex expressions.
        tokenize: Default method for tokenizing the input string. Returns
            lhs, op, and rhs.
        parse_expr: Parse an expression string and generate an Expression.
        to_string: Convert the expression to a string.
        split_into_simple: Split a string into substrings delimited by the
            given delimiter.
        join_expressions: Join a sequence of simple expressions into one
            complex expression string.
        join_strings: Join a sequence of simple expression strings into one
            complex expression string.
    """

    delimiter: str = ""

    def __init__(self, exprstr: str, ctx: c.Context, ops: Sequence[str] = "=<>!+-") -> None:
        """Create Expression with lhs, op, and rhs from an expression string."""
        self.lhs, self.op, self.rhs = self.tokenize(exprstr, ops)

    @staticmethod
    def tokenize(string: str, ops: Sequence[str] = "=<>!+-") -> Tuple[str, str, str]:
        """Tokenize the input string.

        Returns a tuple of strings of the form (lhs, op, rhs).

        Args:
            string: The expression string.
            ops: Valid operator characters.

        Returns:
            A 3-tuple of strings.
        """
        lhs, op, rhs = "", "", ""

        # foo <= bar
        for i, c in enumerate(string):
            if c in ops:
                op = c
                nextc = string[i + 1]
                lhs = string[:i].strip()
                rhs = string[i + 1 :].strip()
                if nextc in ops:  # if
                    op += nextc
                    rhs = rhs[1:].strip()
                break
        return lhs, op, rhs

    @classmethod
    @abstractmethod
    def parse_expr(cls, string: str, ctx: c.Context) -> "Expression":
        """Parse an expression string and generate an Expression."""
        pass

    def to_string(self) -> str:
        """Convert the expression back to a string."""
        return " ".join([self.lhs, self.op, self.rhs])

    @classmethod
    def split_into_simple(cls, complex_str: str) -> List[str]:
        """Split a complex expression into simple expression strings."""
        return complex_str.split(cls.delimiter.strip())

    @classmethod
    def join_expressions(cls, exprs: List[E]) -> str:
        """Join a sequence of simple expressions into one expression string."""
        return cls.join_strings([expr.to_string() for expr in exprs])

    @classmethod
    def join_strings(cls, exprstrs: List[str]) -> str:
        """Join a sequence of simple expressions into one expression string."""
        return cls.delimiter.join(exprstrs)


class UpdateExpression(Expression):
    """Class representing updates when a transition is taken.

    Many UpdateExpressions can be delimited by ", " forming complex
    update expressions. Multiplw updates are executed from left to right.
    For instance, "x = 10, y += x" increments y by 10.

    lhs must be a mutable variable (int). op can be =, +=, or -=.
    rhs can be a constant, a variable, or an int.

    Is extended by ClockResetExpression.
    """

    delimiter = ", "

    @classmethod
    def parse_expr(cls, string: str, ctx: c.Context) -> "UpdateExpression":
        """Construct UpdateExpression with some polymorphism."""
        lhs, _, _ = cls.tokenize(string)
        if ctx.is_clock(lhs):
            return ClockResetExpression(string, ctx)
        return cls(string, ctx)

    def handle_update(self, ctx: c.MutableContext) -> None:
        """Update the mutable context with this expression."""
        cur = ctx.get_val(self.lhs)

        {
            "=": lambda rhs: ctx.set_val(self.lhs, rhs),
            "+=": lambda rhs: ctx.set_val(self.lhs, cur + rhs),
            "-=": lambda rhs: ctx.set_val(self.lhs, cur - rhs),
        }[self.op](ctx.get_val(self.rhs))


class ClockResetExpression(UpdateExpression):
    """Class representing clock resets.

    Extends UpdateExpression with the additional "clock" attribute.
    """

    def __init__(self, exprstr: str, _ctx: c.Context) -> None:
        """Create a ClockResetExpression."""
        super().__init__(exprstr, _ctx)
        self.clock: str = self.lhs


class ConstraintExpression(Expression):
    """Class representing simple constraints.

    Conjunction (&&) of one or more simple constraints constitute a
    complex constraint, used in guards and invariants.

    One of the sides of the expression (lhs or rhs) must INCLUDE either a clock
    or a variable (Note that since clock difference is also supported for clock
    constraints, lhs and rhs should be split further to discover clocks in clock
    difference expressions such as "x - y <= 10".). The other side can be a
    constant, a variable, or an int.  The operator can be <, <=, ==, >=, or >.

    Extends Expression class.
    Extended by ClockConstraintExpression for clock constraints.

    For clock constraints ClockConstraint class should be used. When the type
    of a constaint expression is not known in advance, parse_expr method
    instead of __init__ should be used since parse_expr method automatically
    constructs the ClockConstraint's initializer.
    """

    delimiter = " && "

    @classmethod
    def parse_expr(cls, string: str, ctx: c.Context) -> "ConstraintExpression":
        """Parse an expression string and generate an Expression."""
        lhs, _, rhs = cls.tokenize(string, "<>=")
        for x in lhs.split("-") + rhs.split("-"):
            if ctx.is_clock(x.strip()):
                return ClockConstraintExpression(string, ctx)
        return cls(string, ctx)

    def handle_constraint(self, ctx: c.Context) -> bool:
        """Evaluate the constraint expression based on the current context."""
        return {
            "<": lambda lhs, rhs: lhs < rhs,
            "<=": lambda lhs, rhs: lhs <= rhs,
            "==": lambda lhs, rhs: lhs == rhs,
            ">=": lambda lhs, rhs: lhs >= rhs,
            ">": lambda lhs, rhs: lhs > rhs,
        }[self.op](ctx.get_val(self.lhs), ctx.get_val(self.rhs))


class ClockConstraintExpression(ConstraintExpression):
    """Base class for representing simple clock constraints.

    Attributes:
        clocks: A list of strings with list length either one or two.
            ["x", "y"] denotes x - y and ["x"] simply means x in the simple
            constraint.
        operator: String for the comparison operator. Can be '>', '<', or '='.
        equality: Boolean value for determining the whether the clock value
            can be equal to the threshold value, e.g. x < 10 or x <= 10
        threshold: The value the clock (or clock difference) is constrained by.
            Actually, is a property that "points to" lhs or rhs.
        _threshold_side: Either "left" or "right", used by the threshold property.
    """

    def __init__(self, string: str, ctx: c.Context) -> None:
        """Construct a clock constraint."""
        super().__init__(string, ctx, "<>=")
        self._threshold_side: Union[Literal["left"], Literal["right"]]
        self.clocks: List[str]

        # Determine which side the threshold is.
        if (
            ctx.is_constant(self.lhs)
            or ctx.is_literal(self.lhs)
            or ctx.is_variable(self.lhs)
        ):
            self._threshold_side = "left"
            self.clocks = [c.strip() for c in self.rhs.split("-")]
        else:
            self._threshold_side = "right"
            self.clocks = [c.strip() for c in self.lhs.split("-")]

        self.operator = self.op[0]
        self.equality = len(self.op) == 2

    @property
    def threshold(self):
        if self._threshold_side == "left":
            return self.lhs
        else:
            return self.rhs

    @threshold.setter
    def threshold(self, val):
        if self._threshold_side == "left":
            self.lhs = val
        else:
            self.rhs = val

    def handle_constraint(self, _: c.Context) -> bool:
        """Evaluate the constraint expression based on the current context."""
        raise Exception("ClockConstraints can't be statically checked.")

    def to_string(self, escape=False) -> str:
        """Convert the object to a string.

        If escape is True '<', '>', etc. will be escaped to make the
        resulting string xml-friendly.
        """
        if self.equality:
            self.operator += "="
        res = super().to_string()
        if escape:
            res = (
                res.replace("&", "&amp;")
                .replace('"', "&apos;")
                .replace("<", "&lt;")
                .replace(">", "&gt;")
                .replace("'", "&quot;")
            )
        return res



"""Unit tests for Context and MutableContext."""

from uppaalpy.classes.context import Context, MutableContext


class TestContext:
    """Context tests."""

    @staticmethod
    def test_context_init():
        """Test Context initializer."""
        c = Context(set(), {}, {})
        assert c.clocks == set()
        assert c.constants == {}
        assert c.initial_state == {}

        c = Context(set(["foo", "bar"]), {"x": 13, "y": -1}, {"z": 11})

        assert "foo" in c.clocks and "bar" in c.clocks
        assert c.constants["x"] == 13 and c.constants["y"] == -1
        assert c.initial_state["z"] == 11

    @staticmethod
    def test_context_is_clock():
        """Test is_clock method."""
        c = Context(set(), {}, {})
        assert c.is_clock("notclock") == False

        c = Context(set(["foo"]), {}, {})
        assert c.is_clock("foo")

        c = Context(set(["foo", "bar"]), {}, {})
        assert c.is_clock("foo")
        assert c.is_clock("bar")

        c = Context(set(), {"foo": 3}, {"bar": 4})
        assert c.is_clock("foo") == False
        assert c.is_clock("bar") == False

    @staticmethod
    def test_context_is_constant():
        """Test is_constant method."""
        c = Context(set(), {}, {})
        assert c.is_constant("notconstant") == False

        c = Context(set([]), {"foo": 4}, {})
        assert c.is_constant("foo")

    @staticmethod
    def test_context_is_variable():
        """Test is_variable method."""
        c = Context(set(), {}, {})
        assert c.is_variable("notvar") == False

        c = Context(set([]), {}, {"foo": 4})
        assert c.is_variable("foo")

    @staticmethod
    def test_context_is_literal():
        """Test is_literal method."""
        c = Context(set(), {}, {})
        assert c.is_literal("9")

        c = Context(set([]), {}, {"foo": 4})
        assert c.is_literal("hello") == False

    @staticmethod
    def test_context_get_val():
        """Test get_val method."""
        c = Context(set(), {"foo": 3}, {"bar": 4})
        c.get_val("foo") == 3
        c.get_val("bar") == 4
        c.get_val("15") == 15

    @staticmethod
    def test_context_parse_context():
        """Test parse_context method."""
        c = Context.parse_context(
            """
const int x, y = 10; // comment here.
int a = 11, b;
clock c1, c2,c3;
            """
        )
        print(c.clocks, c.constants, c.initial_state, sep="\n")
        assert c.is_constant("x") and c.get_val("x") == 0
        assert c.is_constant("y") and c.get_val("y") == 10

        assert c.is_variable("a") and c.get_val("a") == 11
        assert c.is_variable("b") and c.get_val("b") == 0

        assert c.is_clock("c1") and c.is_clock("c2") and c.is_clock("c3")
        assert not c.is_clock("c4") and not c.is_clock("x") and not c.is_clock("a")


class TestMutableContext:
    """Tests for the MutableContext class."""

    @staticmethod
    def test_mcontext_to_mutable():
        """Test to_MutableContext method of Context."""
        c = Context(set(), {"foo": 3}, {"bar": 4,})
        mc = c.to_MutableContext()

        assert mc.clocks == set()
        assert mc.get_val("foo") == 3
        assert mc.get_val("bar") == 4

        mc.set_val("bar", 1)

        assert mc.get_val("bar") == 1
        assert c.get_val("foo") == 3
        assert c.get_val("bar") == 4

    @staticmethod
    def test_mcontext_variables_property_get():
        """Test is the property works as inteded."""
        c = MutableContext(set(), {"foo": 3}, {"bar": 4})
        assert c.variables["bar"] == 4

    @staticmethod
    def test_mcontext_variables_property_set():
        """Test is the property works as inteded."""
        c = MutableContext(set(), {"foo": 3}, {"bar": 4})
        c.variables["bar"] = 2
        assert c.variables["bar"] == 2

    @staticmethod
    def test_mcontext_set_val():
        """Test set_val."""
        c = MutableContext(set(), {"foo": 3}, {"bar": 4})
        c.set_val("bar", 10)
        assert c.get_val("bar") == 10



"""Unit test helpers."""
import os

testcase_dir = "lib/uppaalpy/classes/class_tests/"


def list_xml_in_dir(directory):
    """Return list of xml files in dir."""
    return [
        directory.strip("/") + "/" + x
        for x in os.listdir(directory)
        if x.endswith(".xml")
    ]



from uppaalpy.classes.constraint_patcher import (
    ConstraintCache,
    ConstraintChange,
    ConstraintInsert,
    ConstraintPatch,
    ConstraintRemove,
    ConstraintUpdate,
)
from uppaalpy.classes.context import Context, MutableContext
from uppaalpy.classes.expr import (
    ClockConstraintExpression,
    ClockResetExpression,
    ConstraintExpression,
    Expression,
    UpdateExpression,
)
from uppaalpy.classes.nodes import BranchPoint, Location, Node
from uppaalpy.classes.nta import NTA
from uppaalpy.classes.simplethings import (
    ConstraintLabel,
    Declaration,
    Label,
    Name,
    Parameter,
    Query,
    SystemDeclaration,
    UpdateLabel,
)
from uppaalpy.classes.tagraph import TAGraph
from uppaalpy.classes.templates import Template
from uppaalpy.classes.transitions import Nail, Transition
from uppaalpy.path_analysis import *



"""Module for definitions of abstract class Node, and subclasses BranchPoint and Location."""

from typing import Any, Dict, List, Optional, Tuple, Type

import lxml.etree as ET

from uppaalpy.classes import templates as t
from uppaalpy.classes.context import Context
from uppaalpy.classes.expr import ConstraintExpression
from uppaalpy.classes.simplethings import ConstraintLabel, Label, Name


class Node:
    """Abstract class for nodes of the multidigraph in TA templates.

    This class is extended by BranchPoint and Location classes.

    Attributes:
        id: String of the form "idX".
        pos: Pair of ints for storing the position of the node.
    """

    tag = None  # type: Optional[str]
    id = "null"  # type: str
    pos = (0, 0)  # type: Tuple[int, int]

    @staticmethod
    def generate_dict(et, ctx: Context) -> Dict[str, Any]:
        """Construct a dict from an Element object, and return it.

        Notice that only 'id' and 'pos' are relevant for BranchPoints. Other
        attributes are not present in the XML file and ignored for BranchPoint
        objects.
        """
        kw = {}  # type: Dict[str, Any]
        kw["id"] = et.get("id")
        kw["pos"] = int(et.get("x")), int(et.get("y"))
        kw["name"] = Name.from_element(et.find("name"))

        for label in et.iter("label"):
            l_kind = label.get("kind")
            label_obj = Label.from_element(label)

            if l_kind == "invariant":
                label_obj = ConstraintLabel.from_label(label_obj, ctx)

            kw[l_kind] = label_obj

        kw["committed"] = et.find("committed") in et
        kw["urgent"] = et.find("urgent") in et

        return kw

    def to_element(self):
        """Convert this object to an Element. Is extended by Location.to_element."""
        element = ET.Element(
            self.tag,
            attrib={"id": self.id, "x": str(self.pos[0]), "y": str(self.pos[1])},
        )
        return element


class BranchPoint(Node):
    """Derived class of Node.

    The only extension is the added class attribute tag.
    """

    tag = "branchpoint"

    def __init__(self, **kwargs):
        """Accept id string and position pair, and generate a Branchpoint."""
        self.id = kwargs["id"]
        self.pos = kwargs["pos"]

    @classmethod
    def from_element(cls: Type["BranchPoint"], et, ctx: Context) -> "BranchPoint":
        """Generate a dictionary for initialization from et and construct a BP."""
        return cls(**super().generate_dict(et, ctx))


class Location(Node):
    """Derived class of Node.

    This class has additional labels and text fields that are not present in
    BranchPoint objects. Refer to the UPPAAL documentation for information
    on different label kinds.

    Attributes:
        id: id string.
        pos: Int pair.
        name: Name object.
        invariant: ConstraintLabel object for location invariants.
        exponentialrate: Label object. See UPPAAL documentation.
        testcodeEnter: Label object. See UPPAAL documentation.
        testcodeExit: Label object. See UPPAAL documentation.
        comments: Label object for storing comments.
        committed: Boolean value for whether the location is committed.
        urgent: Boolean value for whether the location is urgent.
        template: The parent template, set by TAGraph.
    """

    tag = "location"

    def __init__(self, **kwargs) -> None:
        """Construct a Node from an Element object, and return it.

        This method extends Node.__init__.
        """
        self.id = kwargs["id"]
        self.pos = kwargs["pos"]
        self.name = kwargs.get("name")  # type: Optional[Name]
        self.invariant = kwargs.get("invariant")  # type: Optional[ConstraintLabel]
        self.exponentialrate = kwargs.get("exponentialrate")  # type: Optional[Label]
        self.testcodeEnter = kwargs.get("testcodeEnter")  # type: Optional[Label]
        self.testcodeExit = kwargs.get("testcodeExit")  # type: Optional[Label]
        self.comments = kwargs.get("comments")  # type: Optional[Label]
        self.committed = kwargs.get("committed") or False  # type: bool
        self.urgent = kwargs.get("urgent") or False  # type: bool
        self.template = None  # type: Optional[t.Template]

    @classmethod
    def from_element(cls: Type["Location"], et, ctx: Context) -> "Location":
        """Generate a dictionary for initialization from et and construct a Loc."""
        return cls(**super().generate_dict(et, ctx))

    def to_element(self):
        """Convert this object to an Element.

        This method extends Node.to_element.
        """
        element = super().to_element()
        if self.name is not None:
            element.append(self.name.to_element())
        if self.invariant is not None:
            element.append(self.invariant.to_element())
        if self.exponentialrate is not None:
            element.append(self.exponentialrate.to_element())
        if self.testcodeEnter is not None:
            element.append(self.testcodeEnter.to_element())
        if self.testcodeExit is not None:
            element.append(self.testcodeExit.to_element())
        if self.comments is not None:
            element.append(self.comments.to_element())
        if self.committed:
            element.append(ET.Element("committed"))
        if self.urgent:
            element.append(ET.Element("urgent"))
        return element

    def get_constraints(self) -> List[ConstraintExpression]:
        """Return a list of constraints on this location."""
        if self.invariant is not None:
            return self.invariant.constraints
        else:
            return []

    def get_constraint_label(self) -> Optional[ConstraintLabel]:
        """Return the invariant label."""
        return self.invariant

    def create_constraint_label(self, exp: ConstraintExpression, ctx: Context):
        """Create invariant label."""
        self.invariant = ConstraintLabel("invariant", "", self.pos, ctx, [exp])

    def remove_constraint_label(self) -> None:
        """Remove invariant."""
        self.invariant = None



"""Helper functions for path analysis tests."""
import uppaalpy
import os
from uppaalpy import path_analysis as p

testcase_dir = "lib/uppaalpy/path_tests/testcases/"


def list_nta_in_dir(directory):
    return [
        directory.strip("/") + "/" + x
        for x in os.listdir(directory)
        if x.endswith(".xml")
    ]


def good_nta():
    file_path = testcase_dir + "good_nta"
    return list_nta_in_dir(file_path)


def reachable_nta():
    file_path = testcase_dir + "path_realizable"
    return list_nta_in_dir(file_path)


def not_reachable_nta():
    file_path = testcase_dir + "path_not_realizable"
    return list_nta_in_dir(file_path)


def path_exists():
    file_path = testcase_dir + "path_exists"
    return list_nta_in_dir(file_path)


def path_not_exists():
    file_path = testcase_dir + "path_not_exists"
    return list_nta_in_dir(file_path)


def generator_ntas():
    file_path = "examples/generator"
    return list_nta_in_dir(file_path)

def concatenation_tests():
    file_path = testcase_dir + "path_concatenation"
    return list_nta_in_dir(file_path)


def epsilon_tests():
    file_path = testcase_dir + "epsilon_tests"
    return list_nta_in_dir(file_path)


def read_file(fp):
    return uppaalpy.NTA.from_xml(fp)


def write_to_file(nta):
    return nta.to_file("/tmp/out.xml")


def get_path_from_query_comment(nta):
    path_string = nta.queries[0].comment
    index = int(nta.queries[1].comment)
    res = []
    state = True
    for word in path_string.split():
        if state:
            state = False
            res.append(word)
        else:
            state = True
            res.append(int(word))
    return p.convert_to_path(nta.templates[index], res)


def get_path_concatenation_test(nta):
    ps1, ps2 = nta.queries[0].comment.split("\n")

    index = int(nta.queries[1].comment)

    res_len = int(nta.queries[2].comment)

    res1 = []
    state = True
    for word in ps1.split():
        if state:
            state = False
            res1.append(word)
        else:
            state = True
            res1.append(int(word))

    res2 = []
    state = True
    for word in ps2.split():
        if state:
            state = False
            res2.append(word)
        else:
            state = True
            res2.append(int(word))

    return (
        p.convert_to_path(nta.templates[index], res1),
        p.convert_to_path(nta.templates[index], res2),
        nta.templates[index],
        res_len
    )



"""Tests for NTA class."""

#TODO: reintroduce random changers

from uppaalpy.classes.context import Context
from uppaalpy.classes.nta import NTA
from uppaalpy.classes.simplethings import Declaration, SystemDeclaration

testcase_dir = "lib/uppaalpy/classes/class_tests/"


def _dec_check(l1, l2):
    def sub(l):
        string = (
            l.replace("UTF-8", "utf-8")
            .replace("'", '"')
            .replace("<formula></formula>", "<formula/>")
            .replace("<comment></comment>", "<comment/>")
            # operators
            .replace(" &lt;", "&lt;")
            .replace(" &gt;", "&gt;")
            .replace(" &amp;", "&amp;")
            .replace("&lt; ", "&lt;")
            .replace("&gt; ", "&gt;")
            .replace("&amp; ", "&amp;")
            .replace("= ", "=")
            .replace(" =", "=")
            .replace(" -", "-")
            .replace("- ", "-")
        )
        return string[:-1] if string[-1] == "\n" else string

    def compare_constraints(l1, l2):
        start1 = l1.index(">") + 1  # '>' in ...y="..">
        end1 = l1.index("<", start1)  # '<' in </label>
        c1 = sorted(l1[start1:end1].split("&amp;&amp;"))

        start2 = l2.index(">") + 1  # '>' in ...y="..">
        end2 = l2.index("<", start2)  # '<' in </label>
        c2 = sorted(l2[start2:end2].split("&amp;&amp;"))

        if len(c1) != len(c2):
            return False

        for i in range(len(c1)):
            if c1[i] != c2[i]:
                print("{} and {} are not equal.".format(c1[i], c2[i]))
                return False
        return True

    str1, str2 = l1.strip(), l2.strip()
    if (
        str1.startswith('<label kind="guard"')
        and str2.startswith('<label kind="guard"')
        or str1.startswith('<label kind="invariant"')
        and str2.startswith('<label kind="invariant"')
    ):
        return compare_constraints(sub(l1), sub(l2))

    return sub(l1) == sub(l2)


class TestNTA:
    """Unit tests for NTA."""

    @staticmethod
    def test_nta_init():
        """Test NTA()."""
        nta = NTA(
            declaration=Declaration("foo"),
            templates=[],
            system=SystemDeclaration("bar"),
            queries=[],
            context=Context.empty()
        )
        assert nta.declaration.text == "foo"
        assert nta.system.text == "bar"
        assert nta.templates == []
        assert nta.queries == []
        assert nta._associated_file == ""
        assert nta._doctype == ""

    @staticmethod
    def test_nta_from_xml():
        """Test NTA.from_xml and NTA.from_element."""
        nta = NTA.from_xml(testcase_dir + "nta_xml_files/small_nta.xml")
        assert len(nta.templates) == 2
        assert nta.declaration.text == "// Place global declarations here."
        assert (
            nta.system.text[:-3]
            == "// Place template instantiations here.\nProcess1 = Test1();\nProcess2 = Test2();\n// List one or more processes to be composed into a system.\nsystem Process1, Process2;"
        )
        assert len(nta.queries) == 1

        nta = NTA.from_xml(testcase_dir + "nta_xml_files/big_nta.xml")
        assert len(nta.templates) == 4
        assert nta.declaration.text == "// Place global declarations here.\nchan c1;"
        assert len(nta.queries) == 1

    @staticmethod
    def test_nta_to_file():
        """Test NTA.to_file."""
        path = testcase_dir + "nta_xml_files/small_nta.xml"
        nta = NTA.from_xml(path)
        nta.to_file("/tmp/out.xml")

        with open(path) as inf, open("/tmp/out.xml") as outf:
            inlines, outlines = inf.readlines(), outf.readlines()

        for i in range(len(inlines)):
            assert _dec_check(inlines[i], outlines[i])

        path = testcase_dir + "nta_xml_files/big_nta.xml"
        nta = NTA.from_xml(path)
        nta.to_file("/tmp/out.xml")

        with open(path) as inf, open("/tmp/out.xml") as outf:
            inlines, outlines = inf.readlines(), outf.readlines()

        for i in range(len(inlines)):
            assert _dec_check(inlines[i], outlines[i])

    @staticmethod
    def test_nta_flush_changes_no_changes():
        """Test NTA.flush_constraint_changes() with no changes."""
        path = testcase_dir + "nta_xml_files/small_nta.xml"
        nta = NTA.from_xml(path)
        nta.flush_constraint_changes("/tmp/out.xml")

        with open(path) as inf, open("/tmp/out.xml") as outf:
            inlines, outlines = inf.readlines(), outf.readlines()

        for i in range(len(inlines)):
            assert _dec_check(inlines[i], outlines[i])

        path = testcase_dir + "nta_xml_files/big_nta.xml"
        nta = NTA.from_xml(path)
        nta.flush_constraint_changes("/tmp/out.xml")

        with open(path) as inf, open("/tmp/out.xml") as outf:
            inlines, outlines = inf.readlines(), outf.readlines()

        for i in range(len(inlines)):
            assert _dec_check(inlines[i], outlines[i])

#     @staticmethod
#     def test_nta_flush_changes1():
#         """Test NTA.flush_constraint_changes() with no changes."""
#         path = testcase_dir + "nta_xml_files/small_nta.xml"
#         out1 = "/tmp/out1.xml"
#         out2 = "/tmp/out2.xml"
# 
#         for _ in range(10):
#             nta = random_scenario(path, 0, 0, 10)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 0, 4, 0)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 4, 0, 0)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 2, 2, 2)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 10, 2, 10)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#     @staticmethod
#     def test_nta_flush_changes2():
#         """Test NTA.flush_constraint_changes() with no changes."""
#         path = testcase_dir + "nta_xml_files/big_nta.xml"
#         out1 = "/tmp/out3.xml"
#         out2 = "/tmp/out4.xml"
# 
#         for _ in range(10):
#             nta = random_scenario(path, 0, 0, 10)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 0, 2, 0)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 2, 0, 0)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 2, 2, 2)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 10, 10, 10)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#     @staticmethod
#     def test_nta_flush_changes3():
#         """Test NTA.flush_constraint_changes() with no changes."""
#         path = testcase_dir + "nta_xml_files/test3.xml"
#         out1 = "/tmp/out5.xml"
#         out2 = "/tmp/out6.xml"
# 
#         for _ in range(10):
#             nta = random_scenario(path, 0, 0, 10)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 2, 1, 0)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 2, 0, 0)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 2, 0, 2)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(3):
#             nta = random_scenario(path, 20, 0, 10)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#     @staticmethod
#     def test_nta_flush_changes4():
#         """Test NTA.flush_constraint_changes() with no changes."""
#         path = testcase_dir + "nta_xml_files/test4.xml"
#         out1 = "/tmp/out7.xml"
#         out2 = "/tmp/out8.xml"
# 
#         for _ in range(10):
#             nta = random_scenario(path, 0, 0, 10)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 3, 1, 0)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 2, 0, 0)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 2, 2, 2)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])
# 
#         for _ in range(10):
#             nta = random_scenario(path, 10, 0, 10)
#             nta.flush_constraint_changes(out1)
#             nta.to_file(out2)
# 
#             with open(out1) as flush, open(out2) as write:
#                 inlines, outlines = flush.readlines(), write.readlines()
# 
#             for i in range(len(inlines)):
#                 assert _dec_check(inlines[i], outlines[i])



"""Class definitions for definitions and declarations in the global scope of the TA."""

from copy import deepcopy
from typing import Dict, List, Optional, Set, Tuple, Type

# from uppaalpy.classes import simplethings as s


class Context:
    """An immutable context that is used for variable lookups.

    Extended by MutableContext, which also provides an interface for updating
    variables during a computation.

    Attributes:
        clocks: A list of clock names.
        constants: A dict from identifier strings to values.
        initial_state: A dict from identifier strings to values.

    Methods:
        empty: Create an empty context.
        is_defined: Check if the identifier is defined in the context.
        is_clock: Check if the identifier is a clock.
        is_constant: Check if the identifier is a constant.
        is_variable: Check if the identifier is a variable.
        get_val: Get the corresponding value of an identifier.
        to_MutableContext: Create a copy of the context to be used in other
            computations.
        parse_context: Given a global declaration string, create a context.
    """

    def __init__(
        self,
        clocks: Set[str],
        constants: Dict[str, int],
        initial_state: Dict[str, int],
    ) -> None:
        """Create a context."""
        self.clocks = clocks
        self.constants = constants
        self.initial_state = initial_state

    @classmethod
    def empty(cls: Type["Context"]) -> "Context":
        """Create an empty context."""
        return cls(set(), {}, {})

    def is_defined(self, identifier: str) -> bool:
        """Return True if the identifier is defined in this context."""
        return (
            self.is_clock(identifier)
            or self.is_constant(identifier)
            or self.is_variable(identifier)
        )

    def is_clock(self, identifier: str) -> bool:
        """Return True if the identifier is a clock."""
        return identifier in self.clocks

    def is_constant(self, identifier: str) -> bool:
        """Return True if the identifier is a constant."""
        return self._is_in(identifier, "constants")

    def is_variable(self, identifier: str) -> bool:
        """Return True if the identifier is a variable."""
        return self._is_in(identifier, "initial_state")

    @staticmethod
    def is_literal(string: str) -> bool:
        """Return True if the given string is a number."""
        try:
            int(string)
            return True
        except ValueError:
            return False

    def _is_in(self, identifier: str, attrib: str) -> bool:
        return identifier in getattr(self, attrib).keys()

    def get_val(self, identifier: str) -> int:
        """Return the value of a literal, a constant or a variable.

        Can't be used for clocks.
        """
        if self.is_literal(identifier):
            return int(identifier)
        elif self.is_constant(identifier):
            return self.constants[identifier]
        else:
            return self.initial_state[identifier]

    def to_MutableContext(self) -> "MutableContext":
        """Create a copy of the context to be used in other computations."""
        c = deepcopy(self.clocks)
        const = deepcopy(self.constants)
        init = deepcopy(self.initial_state)
        return MutableContext(c, const, init)

    @classmethod
    def parse_context(cls, declaration: Optional[str]) -> "Context":
        """Get clocks, constants, and variables from global variable.

        Variables of type const int, int, or clock are parsed. Declarations
        should be of form:
        [const] int (identifier "=" value) (identifier "=" value)* ";"
        """
        context = cls.empty()

        if declaration is None:
            return context

        for l in declaration.split("\n"):
            if l.startswith("clock"):
                context._parse_clocks(l)
            elif l.startswith("const int"):
                context._parse_constants(l)
            elif l.startswith("int"):
                context._parse_variables(l)

        return context

    def _parse_clocks(self, line: str) -> None:
        """Given a line starting with "clock" parse clocks."""
        # clock clock1, c2, x; // Some comments...
        #       ^....^  ^^  ^
        for c in line[6 : line.index(";")].split(","):
            self.clocks.add(c.strip())

    def _parse_constants(self, line: str) -> None:
        """Given a line starting with "cont int" parse constants."""
        # const int c = 10, c1 = 100, var; // Some comments...
        #           ^..............,,,,,^
        pairs = self._parse_line(10, line)
        for i, v in pairs:
            self.constants[i] = v

    def _parse_variables(self, line: str) -> None:
        """Given a line starting with "int" parse variables and initial values."""
        # int c = 10, c1 = 100, var; // Some comments...
        #     ^...................^
        pairs = self._parse_line(4, line)
        for i, v in pairs:
            self.initial_state[i] = v

    @staticmethod
    def _parse_line(offset: int, line: str) -> List[Tuple[str, int]]:
        res = []  # type: List[Tuple[str, int]]
        for init in line[offset : line.index(";")].split(","):
            # Declarations with no initialisers are initialized to 0 for ints.
            decl = init.split("=")
            iden = decl[0]
            if len(decl) == 2:
                val = decl[1]
            else:
                val = "0"
            res.append((iden.strip(), int(val.strip())))
        return res


class MutableContext(Context):
    """Mutable version of the Context.

    Properties:
        variables: An alias for the initial_state attribute.

    Methods:
        set_val: Set the value of a variable.
    """

    @property
    def variables(self):
        """Alias for the initial_state attribute."""
        return self.initial_state

    @variables.setter
    def variables(self, x: Dict[str, int]):
        self.initial_state = x

    def set_val(self, identifier: str, val: int):
        """Set the value of a mutable variable."""
        self.variables[identifier] = val



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'lib'}

packages = \
['uppaalpy',
 'uppaalpy.classes',
 'uppaalpy.classes.class_tests',
 'uppaalpy.path_tests']

package_data = \
{'': ['*'],
 'uppaalpy.classes.class_tests': ['constraint_cache_xml_files/*',
                                  'constraint_xml_files/*',
                                  'label_xml_files/*',
                                  'nta_xml_files/*',
                                  'template_xml_files/*'],
 'uppaalpy.path_tests': ['testcases/epsilon_tests/*',
                         'testcases/good_nta/*',
                         'testcases/path_concatenation/*',
                         'testcases/path_exists/*',
                         'testcases/path_not_exists/*',
                         'testcases/path_not_realizable/*',
                         'testcases/path_realizable/*']}

install_requires = \
['lxml>=4.6.2', 'networkx>=2.5', 'ortools>=8.1.8487']

setup_kwargs = {
    'name': 'uppaal-py',
    'version': '0.0.4',
    'description': 'UPPAAL wrapper for Python.',
    'long_description': '# uppaal-py\nPython library for reading, writing, analyzing, and modifying UPPAAL timed automata files. Works with Python >= 3.8.\n\n## Disclaimer\nuppaal-py is a work-in-progress library. For bugs, missing features or documentation please create an issue or send me an email. API is subject to change.\n\n## Dependencies\n* [lxml](https://lxml.de/)\n* [NetworkX](https://github.com/networkx/networkx)\n* [ortools](https://developers.google.com/optimization)\n\n## Installation\nVia pip:\n```\npip install uppaal-py\n```\n\n## License\n[MIT](https://mit-license.org/)\n\n## Features\n- Reading and writing UPPAAL files.\n- LP based path realizability analysis.\n- Finding set of furthest reachable locations with respect to a set of target (unsafe) locations.\n- Working with variables of type `int` and a subset of expressions involving `ints` for guards, invariants, and updates during transitions allowed in UPPAAL in addition to clocks during computations.\n\n## TODO:\n- [ ] Methods for calling UPPAAL/verifyta.\n- [ ] Migrate to [libutap](https://github.com/MASKOR/libutap) for parsing files.\n- [ ] Parameter synthesis for safety property.\n- [ ] [lxml type annotations](https://github.com/lxml/lxml-stubs) and type annotations for the remaining functions.\n- [ ] Auto-generated documentation.\n\n## Non-features\n- Analysis involving network of TA — product of two or more TA can be implemented in the future, though.\n- Symbolic model checking, UPPAAL does that.\n',
    'author': 'Deniz Koluaçık',
    'author_email': 'koluacik@disroot.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/koluacik/uppaal-py',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Unit tests for ConstraintChange and its subclasses."""

import pytest
from pytest_cases.case_parametrizer_new import parametrize_with_cases

from uppaalpy.classes.class_tests.test_constraint_change_cases import (
    CaseCRInit,
    CaseIPatch,
    CaseIinit,
    CaseRFind,
    CaseRPatch,
    CaseUFind,
    CaseUInit,
    CaseUPatch,
)
from uppaalpy.classes.constraint_patcher import (
    ConstraintInsert,
    ConstraintRemove,
)


class TestConstraintRemove:
    @parametrize_with_cases("exp, rem", cases=CaseCRInit)
    def test_init(self, exp, rem):
        if rem == None:
            cr = ConstraintRemove(exp)
            assert cr.remove_label == False
        else:
            cr = ConstraintRemove(exp, remove_label=rem)
            assert cr.remove_label == rem

        assert cr.constraint == exp

    @parametrize_with_cases("exp, lst, res", cases=CaseRFind)
    def test_find(self, exp, lst, res):
        cr = ConstraintRemove(exp)

        if res == -1:
            with pytest.raises(Exception):
                assert cr._find_matching_constraint(lst)
        else:
            assert cr._find_matching_constraint(lst) == res

    @parametrize_with_cases("cr, lines, index, res", cases=CaseRPatch)
    def test_patch(self, cr, lines, index, res):
        cr.patch_line(lines, index)
        assert lines == res


class TestConstraintInsert:
    """Unit tests for ConstraintInsert."""

    @parametrize_with_cases("obj, new_label", cases=CaseIinit)
    def test_constraint_insert_init(self, obj, new_label):
        """Test ConstraintInsert()."""
        ci = ConstraintInsert(obj, new_label)
        assert ci.constraint == obj

    @parametrize_with_cases("ci, lines, res, i, pi", cases=CaseIPatch)
    def test_patch(self, ci, lines, res, i, pi):
        ci.patch_line(lines, i, pi)
        assert len(lines) == len(res)
        for l, r in zip(lines, res):
            assert l == r

class TestConstraintUpdate:
    """Unit tests for ConstraintUpdate."""

    @parametrize_with_cases("c, update, old, new", cases=CaseUInit)
    def test_constraint_update_init(self, c, update, old, new):
        """Test ConstraintUpdate()."""

        assert update.constraint == c
        assert update.old == old
        assert update.new == new

    @parametrize_with_cases("cu, exprs, res", cases=CaseUFind)
    def test_constraint_update_find(self, cu, exprs, res):
        if res is not None:
            assert cu._find_matching_constraint(exprs) == res
        else:
            with pytest.raises(Exception):
                assert cu._find_matching_constraint(exprs)


    @parametrize_with_cases("cu, lines, index, res", cases=CaseUPatch)
    def test_constraint_update_patch(self, cu, lines, index, res):
        cu.patch_line(lines, index)
        assert lines == res



"""Helpers for making random changes on NTA."""
#TODO: Reimplement this module

import random
from uppaalpy import NTA
from uppaalpy.classes.expr import ClockConstraintExpression



def select_random_transition(nta, nonempty=False):
    """Select a random transition.

    If nonempty is True, keep selecting until one with constraints is found.
    """
    while True:
        t = random.choice(nta.templates)
        trans = random.choice(t.graph._transitions)
        if not (nonempty and trans.guard is None):
            break

    return trans


def select_random_location(nta, nonempty=False):
    """Select a random location.

    If nonempty is True, keep selecting until one with constraints is found.
    """
    while True:
        t = random.choice(nta.templates)
        locs = t.graph.get_nodes()
        loc = random.choice(locs)
        if not (nonempty and loc.invariant is None):
            break

    return loc


def make_random_insert(nta):
    """Insert a random constraint to a random transition or location."""
    change_transition = random.choice([True, False])
    ctx = nta.context

    if change_transition:
        trans = select_random_transition(nta)
        nta.change_transition_constraint(
            trans,
            operation="insert",
            simple_constraint=ClockConstraintExpression(
                "c" + random.choice(["<", ">"]) + str(random.randint(1, 100)), ctx
            ),
        )

    else:
        location = select_random_location(nta)
        nta.change_location_constraint(
            location,
            operation="insert",
            simple_constraint=SCConstraint(
                ["x"], random.choice(["<", ">"]), random.randint(1, 100)
            ),
        )


def make_random_remove(nta):
    """Remove a random constraint from a random transition or location."""
    change_transition = random.choice([True, False])

    if change_transition:
        trans = select_random_transition(nta, nonempty=True)
        nta.change_transition_constraint(
            trans,
            operation="remove",
            simple_constraint=random.choice(trans.guard.constraints),
        )

    else:
        location = select_random_location(nta, nonempty=True)
        nta.change_location_constraint(
            location,
            operation="remove",
            simple_constraint=random.choice(location.invariant.constraints),
        )


def make_random_update(nta):
    """Update a random constraint from a random transition or location."""
    change_transition = random.choice([True, False])

    if change_transition:
        trans = select_random_transition(nta, nonempty=True)
        nta.change_transition_constraint(
            trans,
            operation="update",
            simple_constraint=random.choice(trans.guard.constraints),
            threshold_delta=random.randint(1, 10),
        )

    else:
        location = select_random_location(nta, nonempty=True)
        nta.change_location_constraint(
            location,
            operation="update",
            simple_constraint=random.choice(location.invariant.constraints),
            threshold_delta=random.randint(1, 10),
        )


def random_scenario(nta_file, insert_count, remove_count, update_count):
    """Apply random changes for each given change type."""
    nta = NTA.from_xml(nta_file)
    for _ in range(insert_count):
        make_random_insert(nta)

    for _ in range(remove_count):
        make_random_remove(nta)

    for _ in range(update_count):
        make_random_update(nta)

    return nta



"""Test path concatenation."""

import pytest

from uppaalpy.path_analysis import concatenate_paths
from uppaalpy.path_tests.helpers import *


def get_test_data(path):
    nta = read_file(path)
    p1, p2, template, res = get_path_concatenation_test(nta)
    return p1, p2, template, res


@pytest.mark.parametrize("fp", concatenation_tests())
def test_path_concatenation_0(fp):
    p1, p2, t, res = get_test_data(fp)
    paths = concatenate_paths(t, p1, p2)
    assert len(paths) == res
    for p in paths:
        assert len(p) == len(p1) + len(p2) + 1
